from ..detector.detection_probability import PdetGW_compute_cosiota
import pandas as pd
from scipy import stats
from scipy.interpolate import interp2d
import time
import shutil
import os
import numpy as np
import matplotlib.pyplot as plt
from ..cosmology_functions.astrofunction import GW_luminosity_distance
from ..statistic.likelihoods import fake_likelihood_EMGW
import bilby
from ..statistic.likelihoods import log_gaussian_kde_likelihood
from ..statistic.priors import uniform_comoving_volume, z_square, z_uniform
from ..statistic.selection_probability import selection_probability_redshift_cosiota
from scipy.interpolate import interp1d
from tqdm import tqdm


class EMGW_hubble_analysis(object):

    def __init__(self, posterior_samples,
    GW_IFOS, EM_detection_prob=None, d_l_mode='full',
    GW_SNR_thr =12,outdir=time.strftime("%Y_%m_%d_%H_%M_%S"),
    zmax=0.1):

        self.d_l_mode=d_l_mode

        self.posterior_samples = posterior_samples
        #self.GW_IFOS = GW_IFOS
        self.EM_detection_prob=EM_detection_prob
        self.GW_SNR_thr=GW_SNR_thr
        self.outdir=outdir

        if os.path.exists(outdir): # Check if folder is present and delete it
            shutil.rmtree(outdir)
        os.makedirs(outdir)

        self.posterior_kernel_GW=stats.gaussian_kde(np.vstack([posterior_samples['GW']['dL'],posterior_samples['GW']['cosiota']]))
        self.posterior_kernel_EM=stats.gaussian_kde(np.vstack([posterior_samples['EM']['dL'],posterior_samples['EM']['cosiota']]))
        self.posterior_kernel_z=stats.gaussian_kde(np.vstack([posterior_samples['z']['zobs'],posterior_samples['z']['zpec']]))
        self.zmax=zmax

        PdetGW, dcrit_GW=PdetGW_compute_cosiota(detectors = GW_IFOS, SNR_net_thr=GW_SNR_thr, mass_generation='BNS-gaussian')

        if  EM_detection_prob is None:
            PdetEM= lambda d,v:1.
        else:
            print('Computing pdet EM \n')

            df=pd.read_pickle(EM_detection_prob)
            dL = df['dL'][:4000]
            cosi = df['cosi'][:4000]
            d_array= np.linspace(np.min(dL),np.max(dL),400)
            cosiota_array= np.linspace(-1*np.max(cosi),np.max(cosi),400)
            dd, cc = np.meshgrid(d_array,cosiota_array)
            pos = np.stack([dd.flatten(),cc.flatten()])
            dL=np.append(dL,
                 [dL, dL, dL,
                  -dL, -dL, -dL, -dL])
            cosi=np.append(cosi,
                   [-1*cosi,cosi-2.,-cosi+2,
                    cosi,-cosi,cosi-2.,-cosi+2.])
            to_fit = np.vstack([dL,cosi])
            agp_det = stats.gaussian_kde(to_fit)
            values = np.reshape(agp_det(pos).T, dd.shape)
            values = values/np.max(values) # Put pdet to 1 when you detect with Prob 1
            PdetEM=interp2d(d_array,cosiota_array, values, kind='cubic', bounds_error=False, fill_value=0.)


        self.PdetGW=PdetGW
        self.dcrit_GW=dcrit_GW
        self.PdetEM=PdetEM

    def replace_data(self,posterior_samples,outdir):
        self.posterior_samples = posterior_samples
        self.outdir=outdir

        if os.path.exists(outdir): # Check if folder is present and delete it
            shutil.rmtree(outdir)
        os.makedirs(outdir)

        self.posterior_kernel_GW=stats.gaussian_kde(np.vstack([posterior_samples['GW']['dL'],posterior_samples['GW']['cosiota']]))
        self.posterior_kernel_EM=stats.gaussian_kde(np.vstack([posterior_samples['EM']['dL'],posterior_samples['EM']['cosiota']]))
        self.posterior_kernel_z=stats.gaussian_kde(np.vstack([posterior_samples['z']['zobs'],posterior_samples['z']['zpec']]))

    def plot_Pdet(self):

        d_array=np.linspace(0,self.dcrit_GW,100)
        cosiota_array=np.linspace(-1,1,100)
        dd, cc = np.meshgrid(d_array,cosiota_array)


        PdetGW_array=self.PdetGW(d_array,np.arccos(cosiota_array))
        PdetEM_array=self.PdetEM(d_array,cosiota_array)

        if self.EM_detection_prob is not None:
            fig, ax = plt.subplots(1,3,figsize=(15,4),gridspec_kw={'wspace': 0.5})
            im0=ax[0].contourf(dd,cc,PdetGW_array)
            ax[0].colorbar(im0)
            plt.xlabel('Distance [Mpc]')
            plt.ylabel(r'$\cos \iota$')
            plt.title(r'$P_{\rm det}^{\rm GW}$')

            im0=ax[1].contourf(dd,cc,PdetEM_array)
            ax[1].colorbar(im0)
            plt.xlabel('Distance [Mpc]')
            plt.ylabel(r'$\cos \iota$')
            plt.title(r'$P_{\rm det}^{\rm EM}$')


            im0=ax[2].contourf(dd,cc,PdetGW_array*PdetEM_array)
            ax[2].colorbar(im0)
            plt.xlabel('Distance [Mpc]')
            plt.ylabel(r'$\cos \iota$')
            plt.title(r'$P_{\rm det}^{\rm GW} \cdot P_{\rm det}^{\rm EM}$')
            plt.savefig(self.outdir+'/detection_probabilities.pdf')

        else:
            fig, ax = plt.subplots(1,1,figsize=(15,4),gridspec_kw={'wspace': 0.5})
            im0=ax.contourf(dd,cc,PdetGW_array)
            ax.colorbar(im0)
            plt.xlabel('Distance [Mpc]')
            plt.ylabel(r'$\cos \iota$')
            plt.title(r'$P_{\rm det}^{\rm GW}$')

            plt.savefig(self.outdir+'/detection_probabilities.pdf')


    def plot_starting_posteriors(self):

        fig, ax = plt.subplots(1,3,figsize=(15,4),gridspec_kw={'wspace': 0.5})

        for i in range(3):

            if i==0:
                kernel_plot=self.posterior_kernel_GW
                xlab='Distance [Mpc]'
                ylab=r'$\cos \iota$'
                tit='GW  starting posterior'
            elif i==1:
                kernel_plot=self.posterior_kernel_EM
                xlab='Distance [Mpc]'
                ylab=r'$\cos \iota$'
                tit='EM  starting posterior'
            else:
                kernel_plot=self.posterior_kernel_z
                xlab='z observed'
                ylab='z peculiar'
                tit='z starting posterior'

            first_array=kernel_plot.dataset[0,:]
            second_array=kernel_plot.dataset[1,:]

            f_array_grid = np.linspace(np.min(first_array),np.max(first_array),100)
            s_array_grid = np.linspace(np.min(second_array),np.max(second_array),100)

            FF, SS = np.meshgrid(f_array_grid,s_array_grid)
            pos = np.stack([FF.flatten(),SS.flatten()])
            values = np.reshape(kernel_plot(pos).T, FF.shape)

            im0=ax[i].contourf(FF,SS,values)
            ax[i].colorbar(im0)
            plt.xlabel(xlab)
            plt.ylabel(ylab)
            plt.title(tit)

        plt.savefig(self.outdir+'/starting_posteriors.pdf')

    def log_posterior_terms(self,zcosmo,zpec,H0,Om0,cosiota):

        dlt=GW_luminosity_distance(zcosmo,Om0=Om0,H0=H0,mode=self.d_l_mode)
        template=[dlt,cosiota]
        GW_posterior=log_gaussian_kde_likelihood(self.posterior_kernel_GW,template)
        # Assumes flat prior for production
        EM_posterior=log_gaussian_kde_likelihood(self.posterior_kernel_EM,template)
        template=[(1.+zcosmo)*(1.+zpec)-1.,zpec]
        # Assumes flat prior for production
        z_posterior=log_gaussian_kde_likelihood(self.posterior_kernel_z,template)
        return GW_posterior+EM_posterior+z_posterior

    def run_bilby(self,priors,likelihood=fake_likelihood_EMGW,red_prior='uniform_comoving_volume',
    pop_prior='uniform_comoving_volume',prev_analysis=None,**kwargs):
        """This method run bilby using dynesty.

        Parameters
        ----------
        priors: dict
            Dictionary of bilby priors of the parameters. These are *H0*, *Om0*, *alpha_M*, *alpha_T*
            *freq_merger*, *z cosmo*, *zpec*.
        red_prior: str
            Type of prior to use as prior for the redshift measurement. These are *uniform_comoving_volume*
            *z_square*, *z_uniform*
        pop_prior: str
            Type of prior to use as prior for the population and selection effects. These are *uniform_comoving_volume*
            *z_square*, *z_uniform*
        prev_analysis: dict
            a dict containing the 'kernel_posterior_H0' and 'log_evidence'

        Returns
        -------
        results: class
            Bilby result class for the sampler.
        """

        H0_array=np.linspace(priors['H0'].minimum-1,priors['H0'].maximum+1,30)
        Om0=0.308 # Fixes here the Om0


        sel_func=[]
        print('Computing the selection function in H0\n')

        for H0 in tqdm(H0_array):

            eq_redshift_to_dgw= lambda z: GW_luminosity_distance(z,Om0=Om0,H0=H0,mode=self.d_l_mode)

            if pop_prior == 'uniform_comoving_volume':
                redshift_pop_prior = lambda z: uniform_comoving_volume(z,H0=H0,Om0=Om0)
                sel_func.append(selection_probability_redshift_cosiota(self.zmax,self.PdetGW
                ,self.PdetEM,self.dcrit_GW,eq_redshift_to_dgw, redshift_pop_prior,fast_method=False))
            elif pop_prior == 'z_square':
                redshift_pop_prior = lambda z: z_square(z)
                sel_func.append(selection_probability_redshift_cosiota(self.zmax,self.PdetGW
                ,self.PdetEM,self.dcrit_GW,eq_redshift_to_dgw, redshift_pop_prior,fast_method=False))
            elif pop_prior == 'z_uniform':
                redshift_pop_prior = lambda z: z_uniform(z)
                sel_func.append(selection_probability_redshift_cosiota(self.zmax,self.PdetGW
                ,self.PdetEM,self.dcrit_GW,eq_redshift_to_dgw, redshift_pop_prior,fast_method=False))
            elif pop_prior == 'no':
                sel_func.append(0.)

        sel_func = np.array(sel_func)
        log_sel_function_interpolation = interp1d(H0_array,sel_func,kind='cubic',bounds_error=False,fill_value=-np.inf)
        plt.figure()
        plt.plot(H0_array,log_sel_function_interpolation(H0_array))
        plt.xlabel(r'$H_0$')
        plt.ylabel('ln Selection function')
        plt.savefig(self.outdir+'/log_selection_prob.pdf')


        result=bilby.run_sampler(likelihood=fake_likelihood_EMGW(self, red_prior=red_prior, d_l_mode=self.d_l_mode,
        pop_prior=pop_prior,prev_analysis=prev_analysis,log_sel_function_interpolation=log_sel_function_interpolation)
        , priors=priors,outdir=self.outdir, **kwargs)


        return result
