## TO DO
# CHECK THE WAY YOU DO THE ANALYSIS
# TRY AN ANALYSIS WITH SIGMA_z VERY SMALL



import sys
import numpy as np # Simpy nummpy
import pandas as pd
from scipy.interpolate import interp2d, interp1d
from scipy.optimize import fsolve
from tqdm.notebook import tqdm
import copy
import seaborn as sns
from scipy.stats import gaussian_kde

from ..statistic.priors import uniform_comoving_volume,z_square, z_uniform, uniform_comoving_volume_first_order
from ..statistic.selection_probability import selection_probability_redshift_cosiota,selection_probability_redshift
from ..statistic.likelihoods import log_CF_likelihood
from ..cosmology_functions.astrofunction import GW_luminosity_distance
from ..detector.detection_probability import PdetGW_compute_cosiota,detected_binaries_EM, PdetGW_compute,detected_binaries_GW,PdetGW_PdetEM_compute_area
from ..population.binary_generator import generate_binaries
from ..utils.plotting import plot_2d_kde


import matplotlib.pyplot as plt
import matplotlib.cm as cm
from matplotlib import rcParams
rcParams['font.family'] = 'sans-serif'
rcParams['font.sans-serif'] = ['Tahoma']
import pickle


class fast_EM_GW_H0_simulation(object):
    """ This is a class for doing a fast analysis on the Hubble constant inference
    from GW detections with the addition of iota measurementes from the EM counterpart.
    We study the case of BNSs with very accurate localization (for now) and we assume
    the redshift corrected for peculiar motion.

    Parameters
    ----------
    detector_list: list
            List of detector objects of icarogw.
    """


    def __init__(self,detector_list):

        self.detector_list=detector_list

    def generate_binaries(self,H0_injected=70,SNR_thr=12,mode_GW_likelihood='CF'
    ,distance_gen='full',Nbinaries=10000,redshift_gen='uniform_comoving_volume'
    ,mass_generator='BNS-gaussian', outdir='.'):
        """This method initialize the population of binaries and saves it self.binaries_dict

        Parameters
        ----------
        H0_injected: float
            Hubble constant of your universe in km Mpc-1 s-1
        SNR_thr: float
            SNR threshold for the detection of the GW
        mode_GW_likelihood: string
            This is the mode that you use for the GW likelihood. For now only 'CF'
            is implemented
        distance_gen: str
            This is a string saying what d_L -> z model you want. You can use 'full', complete Friedmann
            or 'linear' for c*z/H0, 'first_order'
        Nbinaries: int
            Th number of binaries to simulate
        redshift_gen: str
            What distribution the binaries has in redshift. Possible 'uniform_comoving_volume','z_sqaure','z_uniform'
        mass_generator: str
            Binary neutron stars by defauls
        outdir:str
            Where you want to save this particular population and its plots
        """

        print('We are generating binaries. These will be saved in the attribute binaries dict')
        self.outdir=outdir
        self.H0_injected = H0_injected
        self.SNR_thr=SNR_thr
        self.mode_GW_likelihood=mode_GW_likelihood
        self.distance_gen=distance_gen
        self.redshift_gen=redshift_gen
        self.mass_generator=mass_generator
        self.Nbinaries=Nbinaries

        if distance_gen=='full':
            # This is the function for generating distances in cosmology
            distance_generator = lambda z : GW_luminosity_distance(z,H0=H0_injected,mode='full')
        elif distance_gen=='linear':
            distance_generator = lambda z : GW_luminosity_distance(z,H0=H0_injected,mode='linear')
        elif distance_gen=='first_order':
            distance_generator = lambda z : GW_luminosity_distance(z,H0=H0_injected,mode='first_order')
        else:
            sys.exit('Not yet implemented')

        if redshift_gen=='uniform_comoving_volume':
            # This is the distribution in redshift of the binaries
            redshift_generator = lambda z : np.exp(uniform_comoving_volume(z,H0=H0_injected))
        elif redshift_gen=='z_square':
            redshift_generator = lambda z : np.exp(z_square(z))
        elif redshift_gen=='z_uniform':
            redshift_generator = lambda z : np.exp(z_uniform(z))
        elif redshift_gen=='uniform_comoving_volume_first_order':
            # This is the distribution in redshift of the binaries
            redshift_generator = lambda z : np.exp(uniform_comoving_volume_first_order(z,H0=H0_injected))
        else:
            sys.exit('Not yet implemented')

        self.PdetGW, self.dcrit = PdetGW_compute(False,horizon = None, detectors = self.detector_list
        , SNR_net_thr=self.SNR_thr,mass_generation=self.mass_generator,mode=self.mode_GW_likelihood)

        # Below we compute the maximum redshift at which GW will be detectable
        to_solve = lambda z: distance_generator(z)-self.dcrit
        self.zmax = fsolve(to_solve,0)[0]
        print('For this cosmology. Your zmax is {:.4f}'.format(self.zmax))
        self.binaries_dict=generate_binaries(Nbinaries,redshift_generator,
        zmax=self.zmax,mass_generation=self.mass_generator)

    def compute_GW_detected_binaries(self):
        """ This method adds a field to self.binaries_dict with the optimal
        GW SNR, the detected SNR and if a binary is detected or not.
        """

        print('Computing the detected GW SNR and lum distance and cosiota\n')
        if self.distance_gen=='full':
            # This is the function for generating distances in cosmology
            distance_generator = lambda zz : GW_luminosity_distance(zz,H0=self.H0_injected,mode='full')
        elif self.distance_gen=='linear':
            distance_generator = lambda zz : GW_luminosity_distance(zz,H0=self.H0_injected,mode='linear')
        elif self.distance_gen=='first_order':
            distance_generator = lambda zz : GW_luminosity_distance(zz,H0=self.H0_injected,mode='first_order')
        else:
            sys.exit('Not yet implemented')

        self.binaries_dict=detected_binaries_GW(self.binaries_dict,
        distance_generator,self.detector_list,SNR_thr=self.SNR_thr,mode=self.mode_GW_likelihood)

    def compute_EM_detected_binaries(self,EM_Pdet_file,true_value_mode=True,sigma_d_EM=0,sigma_iota_EM=0,fow=np.inf):
        """ This method computes which binaries are detected in EM. It adds a field 'EM_detected'
        to self.binaries_dict

        Parameters
        ----------
        EM_Pdet_file: str
            Raph's file with Pdet_EM
        true_value_mode: bool
            If true, when deciding if a binary is detected, the code assumes a perfect
            measurement of cosiota and d_l
        sigma_d_EM: float
            The sigma in Mpc for the EM likelihood. NB this does not influece the detection criteria
            if true_value_mode is true.
        sigma_iota_EM: float
            The sigma iota in radians for the EM likelihood. NB this does not influece the detection criteria
            if true_value_mode is true.
        """
        print('Computing the detected EM and lum distance and cosiota\n')

        if self.distance_gen=='full':
            # This is the function for generating distances in cosmology
            distance_generator = lambda zz : GW_luminosity_distance(zz,H0=self.H0_injected,mode='full')
        elif self.distance_gen=='linear':
            distance_generator = lambda zz : GW_luminosity_distance(zz,H0=self.H0_injected,mode='linear')
        elif self.distance_gen=='first_order':
            distance_generator = lambda zz : GW_luminosity_distance(zz,H0=self.H0_injected,mode='first_order')
        else:
            sys.exit('Not yet implemented')

        self.EM_Pdet_file=EM_Pdet_file
        self.sigma_iota_EM=sigma_iota_EM
        self.sigma_d_EM=sigma_d_EM
        self.fow=fow

        localization_area = 10.0**(5.06)/self.binaries_dict['SNR_detected']**2 # 90% localization of the event in deg^2
        self.binaries_dict['localization_area'] = localization_area
        self.binaries_dict['sky_area_detected']= localization_area<=fow


        self.binaries_dict=detected_binaries_EM(self.binaries_dict,
        distance_generator,sigma_d_EM,sigma_iota_EM,EM_Pdet=EM_Pdet_file,true_value_mode=true_value_mode)


    def compute_z_detected_binaries(self,z_Pdet_file=None,sigma_z=150):
        """
        Compute the detected binaries in z and z_detected
        """

        self.sigma_z=sigma_z
        c=299792.458 # Speed of light in km/s

        if z_Pdet_file is not None:
            sys.exit('Not yet implemented')
        else:
            z_obs_out = self.binaries_dict['z']+np.random.randn(len(self.binaries_dict['tgps']))*(sigma_z/c)
            self.binaries_dict['z_detected']=z_obs_out

    def compute_and_Pdet(self):
        """
        Plots Pdets
        """
        print('I am computing the Pdet sky area')
        PdetGWEM_area, dcrit_area=  PdetGW_PdetEM_compute_area(self.EM_Pdet_file,
        self.detector_list, self.fow, SNR_net_thr=self.SNR_thr, Nsample=self.Nbinaries,
        mass_generation=self.mass_generator)

        self.PdetGWEM_area=PdetGWEM_area
        self.dcrit_area=dcrit_area

        print('computing PdetGW (d, iota)')
        PdetGW_dc, dcrit_GW=PdetGW_compute_cosiota(detectors = self.detector_list
        , SNR_net_thr=self.SNR_thr, Nsample=self.Nbinaries, mass_generation=self.mass_generator)

        self.dcrit_GW=dcrit_GW

        self.PdetGW_dc=PdetGW_dc

        print('Computing pdetEM (d, cosiota iota) \n')

        df=pd.read_pickle(self.EM_Pdet_file)
        dL = df['dL'][:4000]
        cosi = df['cosi'][:4000]
        d_array= np.linspace(np.min(dL),np.max(dL),400)
        cosiota_array= np.linspace(-1*np.max(cosi),np.max(cosi),400)
        dd, cc = np.meshgrid(d_array,cosiota_array)
        pos = np.stack([dd.flatten(),cc.flatten()])
        dL=np.append(dL,
             [dL, dL, dL,
              -dL, -dL, -dL, -dL])
        cosi=np.append(cosi,
               [-1*cosi,cosi-2.,-cosi+2,
                cosi,-cosi,cosi-2.,-cosi+2.])
        to_fit = np.vstack([dL,cosi])
        agp_det = gaussian_kde(to_fit)
        values = np.reshape(agp_det(pos).T, dd.shape)
        values = values/np.max(values) # Put pdet to 1 when you detect with Prob 1
        self.PdetEM=interp2d(d_array,cosiota_array, values, kind='cubic', bounds_error=False, fill_value=0.)
        self.dmax=np.max([self.dcrit,self.dcrit_GW,self.dcrit_area])

    def plot_Pdet(self):

        d_array=np.linspace(0,self.dmax,100)
        cosiota_array=np.linspace(-1,1,100)
        dd, cc = np.meshgrid(d_array,cosiota_array)

        PdetGW_array=self.PdetGW_dc(d_array,np.arccos(cosiota_array))
        PdetEM_array=self.PdetEM(d_array,cosiota_array)
        Pdet_foe=self.PdetGWEM_area(d_array)

        fig, ax = plt.subplots(2,2,figsize=(10,10),gridspec_kw={'wspace': 0.5})
        im0=ax[0,0].contourf(dd,cc,PdetGW_array)
        ax[0,0].colorbar(im0)
        plt.xlabel('Distance [Mpc]')
        plt.ylabel(r'$\cos \iota$')
        plt.title(r'$P_{\rm det}^{\rm GW}$')

        im0=ax[0,1].contourf(dd,cc,PdetEM_array)
        ax[0,1].colorbar(im0)
        plt.xlabel('Distance [Mpc]')
        plt.ylabel(r'$\cos \iota$')
        plt.title(r'$P_{\rm det}^{\rm EM}$')


        im0=ax[1,0].contourf(dd,cc,PdetGW_array*PdetEM_array)
        ax[1,0].colorbar(im0)
        plt.xlabel('Distance [Mpc]')
        plt.ylabel(r'$\cos \iota$')
        plt.title(r'$P_{\rm det}^{\rm GW} \cdot P_{\rm det}^{\rm EM}$')

        plot_joint=np.sum(PdetGW_array*PdetEM_array,axis=0)/len(cosiota_array)
        plot_GW=np.sum(PdetGW_array,axis=0)/len(cosiota_array)
        plot_EM=np.sum(PdetEM_array,axis=0)/len(cosiota_array)

        ax[1,1].plot(d_array,Pdet_foe,label='GW+EM fow')
        ax[1,1].plot(d_array,plot_joint,label='GW+EM')
        ax[1,1].plot(d_array,plot_GW,label='GW')
        ax[1,1].plot(d_array,plot_EM,label='EM')
        ax[1,1].legend()
        ax[1,1].set_xlabel('Distance [Mpc]')
        ax[1,1].set_ylabel('Detection prob')
        plt.savefig(self.outdir+'/detection_probabilities.pdf')

    def compute_selection_effect(self,H0_min,H0_max):
        """Computes and interpolates the selection effect in H0.

        Parameters
        ----------
        H0_min: float
            In km Mpc^-1 s^-1
        H0_max: float
            In km Mpc^-1 s^-1
        """

        self.H0_min=H0_min
        self.H0_max=H0_max
        H0_array=np.linspace(H0_min-1,H0_max+1,30)

        sel_func=[]
        sel_func_GW=[]
        sel_func_area=[]

        print('Computing the selection function in H0\n')

        for H0 in tqdm(H0_array):

            eq_redshift_to_dgw= lambda z: GW_luminosity_distance(z,H0=H0,mode=self.distance_gen)
            if self.redshift_gen == 'uniform_comoving_volume':
                redshift_pop_prior = lambda z: uniform_comoving_volume(z,H0=H0)
                sel_func.append(selection_probability_redshift_cosiota(4*self.zmax,self.PdetGW_dc
                ,self.PdetEM,self.dmax,eq_redshift_to_dgw, redshift_pop_prior,fast_method=False))
            elif self.redshift_gen == 'z_square':
                redshift_pop_prior = lambda z: z_square(z)
                sel_func.append(selection_probability_redshift_cosiota(4*self.zmax,self.PdetGW_dc
                ,self.PdetEM,self.dmax,eq_redshift_to_dgw, redshift_pop_prior,fast_method=False))
            elif self.redshift_gen == 'z_uniform':
                redshift_pop_prior = lambda z: z_uniform(z)
                sel_func.append(selection_probability_redshift_cosiota(4*self.zmax,self.PdetGW_dc
                ,self.PdetEM,self.dmax,eq_redshift_to_dgw, redshift_pop_prior,fast_method=False))
            elif self.redshift_gen == 'uniform_comoving_volume_first_order':
                redshift_pop_prior = lambda z: uniform_comoving_volume_first_order(z,H0=H0)
                sel_func.append(selection_probability_redshift_cosiota(4*self.zmax,self.PdetGW_dc
                ,self.PdetEM,self.dmax,eq_redshift_to_dgw, redshift_pop_prior,fast_method=False))
            else:
                sys.exit('not yet implemented')

            if self.redshift_gen == 'uniform_comoving_volume':
                redshift_pop_prior = lambda z: uniform_comoving_volume(z,H0=H0)
                sel_func_GW.append(selection_probability_redshift(4*self.zmax,self.PdetGW
                ,self.dmax,eq_redshift_to_dgw, redshift_pop_prior))
            elif self.redshift_gen == 'z_square':
                redshift_pop_prior = lambda z: z_square(z)
                sel_func_GW.append(selection_probability_redshift(4*self.zmax,self.PdetGW
                ,self.dmax,eq_redshift_to_dgw, redshift_pop_prior))
            elif self.redshift_gen == 'z_uniform':
                redshift_pop_prior = lambda z: z_uniform(z)
                sel_func_GW.append(selection_probability_redshift(4*self.zmax,self.PdetGW
                ,self.dmax,eq_redshift_to_dgw, redshift_pop_prior))
            elif self.redshift_gen == 'uniform_comoving_volume_first_order':
                redshift_pop_prior = lambda z: uniform_comoving_volume_first_order(z,H0=H0)
                sel_func_GW.append(selection_probability_redshift(4*self.zmax,self.PdetGW
                ,self.dmax,eq_redshift_to_dgw, redshift_pop_prior))
            else:
                sys.exit('not yet implemented')

            if self.redshift_gen == 'uniform_comoving_volume':
                redshift_pop_prior = lambda z: uniform_comoving_volume(z,H0=H0)
                sel_func_area.append(selection_probability_redshift(4*self.zmax,self.PdetGWEM_area
                ,self.dmax,eq_redshift_to_dgw, redshift_pop_prior))
            elif self.redshift_gen == 'z_square':
                redshift_pop_prior = lambda z: z_square(z)
                sel_func_area.append(selection_probability_redshift(4*self.zmax,self.PdetGWEM_area
                ,self.dmax,eq_redshift_to_dgw, redshift_pop_prior))
            elif self.redshift_gen == 'z_uniform':
                redshift_pop_prior = lambda z: z_uniform(z)
                sel_func_area.append(selection_probability_redshift(4*self.zmax,self.PdetGWEM_area
                ,self.dmax,eq_redshift_to_dgw, redshift_pop_prior))
            if self.redshift_gen == 'uniform_comoving_volume_first_order':
                redshift_pop_prior = lambda z: uniform_comoving_volume_first_order(z,H0=H0)
                sel_func_area.append(selection_probability_redshift(4*self.zmax,self.PdetGWEM_area
                ,self.dmax,eq_redshift_to_dgw, redshift_pop_prior))
            else:
                sys.exit('not yet implemented')

        sel_func = np.array(sel_func)
        sel_func_GW = np.array(sel_func_GW)
        sel_func_area = np.array(sel_func_area)

        log_sel_function_interpolation = interp1d(H0_array,sel_func-sel_func[0],kind='cubic',bounds_error=False,fill_value=-np.inf)
        log_sel_function_interpolation_GW = interp1d(H0_array,sel_func_GW-sel_func_GW[0],kind='cubic',bounds_error=False,fill_value=-np.inf)
        log_sel_function_interpolation_area = interp1d(H0_array,sel_func_area-sel_func_area[0],kind='cubic',bounds_error=False,fill_value=-np.inf)

        plt.figure()
        plt.plot(H0_array,log_sel_function_interpolation(H0_array),label='GW+EM')
        plt.plot(H0_array,log_sel_function_interpolation_GW(H0_array),label='GW')
        plt.plot(H0_array,log_sel_function_interpolation_area(H0_array),label='GW+EM fow')

        plt.xlabel(r'$H_0$')
        plt.ylabel('ln Selection function')
        plt.legend()
        plt.savefig(self.outdir+'log_selection_prob.pdf')

        self.log_sel_function_interpolation=log_sel_function_interpolation
        self.log_sel_function_interpolation_GW=log_sel_function_interpolation_GW
        self.log_sel_function_interpolation_area=log_sel_function_interpolation_area

    def run_analysis(self,Ndetections,H0_posterior=np.linspace(40,120,100),permutation=False,
    Nsamp=5000,injections='GW_EM',only_GW=False):

        dH0=H0_posterior[1]-H0_posterior[0]
        c=299792.458 # Speed of light in km/s
        if isinstance(injections, str):
            if injections=='GW':
                ilabel_GW_detected = np.where(self.binaries_dict['GW_detected'])[0]
                print('You have detected {:d} GWs\n'.format(len(ilabel_GW_detected)))
            elif injections=='GW_EM_fow':
                ilabel_GW_detected = np.where(self.binaries_dict['EM_detected'] & self.binaries_dict['GW_detected']
                 & self.binaries_dict['sky_area_detected'])[0]
                print('FOW: You have detected {:d} GWs and EM together \n'.format(len(ilabel_GW_detected)))
            elif injections=='GW_EM':
                ilabel_GW_detected = np.where(self.binaries_dict['EM_detected'] & self.binaries_dict['GW_detected'])[0]
                print('You have detected {:d} GWs and EM together \n'.format(len(ilabel_GW_detected)))
        else:
            ilabel_GW_detected=injections
        # Permute the order of detections if flag is given.
        if permutation:
            ilabel_GW_detected=np.random.permutation(ilabel_GW_detected)
        num_inj=Ndetections
        pdf_combined=np.ones_like(H0_posterior)
        i_inj=0
        CF = log_CF_likelihood(self.detector_list)
        pdf_single_out=[]

        pdf_combined=np.zeros_like(H0_posterior)
        for i in tqdm(ilabel_GW_detected, desc='Calculating H0 posterior'):
            pdf_single=np.zeros_like(H0_posterior)
            z_samples = self.binaries_dict['z_detected'][i]+np.random.randn(Nsamp)*(self.sigma_z/c)
            signal={'tgps': np.array([self.binaries_dict['tgps'][i]]),
                'dec': np.array([self.binaries_dict['dec'][i]]),
                'RA': np.array([self.binaries_dict['RA'][i]]),
                'psi': np.array([self.binaries_dict['psi'][i]]),
                'iota': np.array([self.binaries_dict['det_iota'][i]]),
                'SNR_detected': np.array([self.binaries_dict['SNR_detected'][i]])}
            if isinstance(injections, str):

                if (((injections=='GW_EM') | (injections=='GW_EM_fow')) & (not only_GW)):
                    iota_a=np.random.randn(Nsamp)*self.sigma_iota_EM+self.binaries_dict['det_iota_EM'][i]
                    c_array = np.cos(iota_a)
                    for j in range(len(H0_posterior)):
                        eq_redshift_to_dgw= lambda z: GW_luminosity_distance(z,H0=H0_posterior[j],mode=self.distance_gen)
                        if self.redshift_gen == 'uniform_comoving_volume':
                            redshift_pop_prior = lambda z: np.exp(uniform_comoving_volume(z,H0=H0_posterior[j]))
                        elif self.redshift_gen == 'z_square':
                            redshift_pop_prior = lambda z: np.exp(z_square(z))
                        elif self.redshift_gen == 'z_uniform':
                            redshift_pop_prior = lambda z: np.exp(z_uniform(z))
                        elif self.redshift_gen == 'uniform_comoving_volume_first_order':
                            redshift_pop_prior = lambda z: np.exp(uniform_comoving_volume_first_order(z,H0=H0_posterior[j]))
                        else:
                            sys.exit('not yet implemented')

                        if self.redshift_gen == 'uniform_comoving_volume':
                            prior=np.array([redshift_pop_prior(zz) for zz in z_samples])
                        else:
                            prior=redshift_pop_prior(z_samples)

                        if self.distance_gen == 'full':
                            dist_vals=np.array([eq_redshift_to_dgw(zz) for zz in z_samples])
                        else:
                            dist_vals=eq_redshift_to_dgw(z_samples)

                        dist_vals/=self.binaries_dict['det_luminosity_distance'][i]
                        pdf_single[j]=np.sum(np.exp(CF.log_pdf_point(dist_vals,c_array,signal))*prior*np.abs(np.sin(iota_a)))
                else:
                    c_array = np.random.rand(Nsamp)*2-1
                    for j in range(len(H0_posterior)):
                        eq_redshift_to_dgw= lambda z: GW_luminosity_distance(z,H0=H0_posterior[j],mode=self.distance_gen)
                        if self.redshift_gen == 'uniform_comoving_volume':
                            redshift_pop_prior = lambda z: np.exp(uniform_comoving_volume(z,H0=H0_posterior[j]))
                        elif self.redshift_gen == 'z_square':
                            redshift_pop_prior = lambda z: np.exp(z_square(z))
                        elif self.redshift_gen == 'z_uniform':
                            redshift_pop_prior = lambda z: np.exp(z_uniform(z))
                        elif self.redshift_gen == 'uniform_comoving_volume_first_order':
                            redshift_pop_prior = lambda z: np.exp(uniform_comoving_volume_first_order(z,H0=H0_posterior[j]))
                        else:
                            sys.exit('not yet implemented')

                        if self.redshift_gen == 'uniform_comoving_volume':
                            prior=np.array([redshift_pop_prior(zz) for zz in z_samples])
                        else:
                            prior=redshift_pop_prior(z_samples)

                        if self.distance_gen == 'full':
                            dist_vals=np.array([eq_redshift_to_dgw(zz) for zz in z_samples])
                        else:
                            dist_vals=eq_redshift_to_dgw(z_samples)

                        dist_vals/=self.binaries_dict['det_luminosity_distance'][i]
                        pdf_single[j]=np.sum(np.exp(CF.log_pdf_point(dist_vals,c_array,signal))*prior)
            else:

                if ((self.binaries_dict['GW_detected'][i] & self.binaries_dict['EM_detected'][i] &
                self.binaries_dict['sky_area_detected'][i]) & (not only_GW)):
                    print('USING EM INFO \n')
                    iota_a=np.random.randn(Nsamp)*self.sigma_iota_EM+self.binaries_dict['det_iota_EM'][i]
                    c_array = np.cos(iota_a)
                    for j in range(len(H0_posterior)):
                        eq_redshift_to_dgw= lambda z: GW_luminosity_distance(z,H0=H0_posterior[j],mode=self.distance_gen)
                        if self.redshift_gen == 'uniform_comoving_volume':
                            redshift_pop_prior = lambda z: np.exp(uniform_comoving_volume(z,H0=H0_posterior[j]))
                        elif self.redshift_gen == 'z_square':
                            redshift_pop_prior = lambda z: np.exp(z_square(z))
                        elif self.redshift_gen == 'z_uniform':
                            redshift_pop_prior = lambda z: np.exp(z_uniform(z))
                        elif self.redshift_gen == 'uniform_comoving_volume_first_order':
                            redshift_pop_prior = lambda z: np.exp(uniform_comoving_volume_first_order(z,H0=H0_posterior[j]))
                        else:
                            sys.exit('not yet implemented')

                        if self.redshift_gen == 'uniform_comoving_volume':
                            prior=np.array([redshift_pop_prior(zz) for zz in z_samples])
                        else:
                            prior=redshift_pop_prior(z_samples)

                        if self.distance_gen == 'full':
                            dist_vals=np.array([eq_redshift_to_dgw(zz) for zz in z_samples])
                        else:
                            dist_vals=eq_redshift_to_dgw(z_samples)

                        dist_vals/=self.binaries_dict['det_luminosity_distance'][i]
                        pdf_single[j]=np.sum(np.exp(CF.log_pdf_point(dist_vals,c_array,signal))*prior*np.abs(np.sin(iota_a)))
                else:
                    c_array = np.random.rand(Nsamp)*2-1
                    for j in range(len(H0_posterior)):
                        eq_redshift_to_dgw= lambda z: GW_luminosity_distance(z,H0=H0_posterior[j],mode=self.distance_gen)
                        if self.redshift_gen == 'uniform_comoving_volume':
                            redshift_pop_prior = lambda z: np.exp(uniform_comoving_volume(z,H0=H0_posterior[j]))
                        elif self.redshift_gen == 'z_square':
                            redshift_pop_prior = lambda z: np.exp(z_square(z))
                        elif self.redshift_gen == 'z_uniform':
                            redshift_pop_prior = lambda z: np.exp(z_uniform(z))
                        elif self.redshift_gen == 'uniform_comoving_volume_first_order':
                            redshift_pop_prior = lambda z: np.exp(uniform_comoving_volume_first_order(z,H0=H0_posterior[j]))
                        else:
                            sys.exit('not yet implemented')

                        if self.redshift_gen == 'uniform_comoving_volume':
                            prior=np.array([redshift_pop_prior(zz) for zz in z_samples])
                        else:
                            prior=redshift_pop_prior(z_samples)

                        if self.distance_gen == 'full':
                            dist_vals=np.array([eq_redshift_to_dgw(zz) for zz in z_samples])
                        else:
                            dist_vals=eq_redshift_to_dgw(z_samples)

                        dist_vals/=self.binaries_dict['det_luminosity_distance'][i]
                        pdf_single[j]=np.sum(np.exp(CF.log_pdf_point(dist_vals,c_array,signal))*prior)

            pdf_single=np.log(pdf_single/np.sum(pdf_single*dH0))
            if isinstance(injections, str):
                if (injections=='GW_EM') & (not only_GW):
                    pdf_single-=self.log_sel_function_interpolation(H0_posterior)
                elif (injections=='GW_EM') & (only_GW):
                    pdf_single-=self.log_sel_function_interpolation_GW(H0_posterior)
                elif (injections=='GW_EM_fow') & (not only_GW):
                    pdf_single-=self.log_sel_function_interpolation_area(H0_posterior)
                elif (injections=='GW_EM_fow') & (only_GW):
                    pdf_single-=self.log_sel_function_interpolation_GW(H0_posterior)
            else:
                if only_GW:
                    pdf_single-=self.log_sel_function_interpolation_GW(H0_posterior)
                elif (self.binaries_dict['GW_detected'][i] &
                self.binaries_dict['EM_detected'][i] &
                self.binaries_dict['sky_area_detected'][i]):
                    pdf_single-=self.log_sel_function_interpolation_area(H0_posterior)
                elif (self.binaries_dict['GW_detected'][i] &
                self.binaries_dict['EM_detected'][i]):
                    pdf_single-=self.log_sel_function_interpolation(H0_posterior)
                else:
                    pdf_single-=self.log_sel_function_interpolation_GW(H0_posterior)

            pdf_single-=np.log(np.sum(np.exp(pdf_single)*dH0))
            pdf_single_out.append(pdf_single)
            pdf_combined+=pdf_single
            pdf_combined-=np.log(np.sum(np.exp(pdf_combined)*dH0))
            i_inj +=1

            if i_inj==num_inj:
                break


        return pdf_single_out,pdf_combined


    def run_anaysis(self,Ndetections,H0_posterior=np.linspace(40,120,100),permutation=False,
    Nsamp_GW=10000,Nsamp_EM=5000,Nsamp_comb=10000,Nsamp_z=500,injections='GW_EM',only_GW=False):
        """ Runs the analysis for H0 and provides the single posterior distributions and the combined one

        Parameters
        ----------
        Ndetections: int
            Number of detections to stuck
        H0_posterior: np.array
            Array of the H0 grid
        permutation: boolean
            If true, permutes the detected binaries and pick up the first Ndetections
        Nsamp_GW: int
            Number of samples to draw from CF
        Nsamp_EM: int
            Number of samples to draw from EM likelihood
        Nsamp_comb: int
            Number of samples to draw from the GW+EM likelihood
        Nsamp_z: int
            Number of samples in redshift. The cost rapidly increase with this.
        injections: str or np.array
            Select the injections you want to stuy 'GW' or 'GW_EM'. Otherwise provide
            ID of the injection
        only_GW: bool
            If true, use only the GW likelihood

        Returns
        -------
        pdf_single_out: list
            list containing the gaussian_kde of the single posteirors
        pdf_combined: np.array
            Final combined posterior on the grid
        """
        c=299792.458 # Speed of light in km/s
        dH0=H0_posterior[1]-H0_posterior[0]
        if injections=='GW':
            ilabel_GW_detected = np.where(self.binaries_dict['GW_detected'])[0]
            print('You have detected {:d} GWs\n'.format(len(ilabel_GW_detected)))
        elif injections=='GW_EM_fow':
            ilabel_GW_detected = np.where(self.binaries_dict['EM_detected'] & self.binaries_dict['GW_detected']
             & self.binaries_dict['sky_area_detected'])[0]
            print('FOW: You have detected {:d} GWs and EM together \n'.format(len(ilabel_GW_detected)))
        elif injections=='GW_EM':
            ilabel_GW_detected = np.where(self.binaries_dict['EM_detected'] & self.binaries_dict['GW_detected'])[0]
            print('You have detected {:d} GWs and EM together \n'.format(len(ilabel_GW_detected)))
        else:
            ilabel_GW_detected=injections


        # Permute the order of detections if flag is given.
        if permutation:
            ilabel_GW_detected=np.random.permutation(ilabel_GW_detected)
        num_inj=Ndetections
        pdf_combined=np.ones_like(H0_posterior)
        i_inj=0

        CF = log_CF_likelihood(self.detector_list)

        pdf_single_out=[]
        pdf_combined=np.zeros_like(H0_posterior)
        for i in ilabel_GW_detected:
            print('Injection-'+str(i_inj+1))


            z_samples = self.binaries_dict['z_detected'][i]+np.random.randn(Nsamp_z)*(self.sigma_z/c)

            signal={'tgps': np.array([self.binaries_dict['tgps'][i]]),
                'dec': np.array([self.binaries_dict['dec'][i]]),
                'RA': np.array([self.binaries_dict['RA'][i]]),
                'psi': np.array([self.binaries_dict['psi'][i]]),
                'iota': np.array([self.binaries_dict['det_iota'][i]]),
                'SNR_detected': np.array([self.binaries_dict['SNR_detected'][i]])}


            if isinstance(injections, str):
                if (((injections=='GW_EM') | (injections=='GW_EM_fow')) & (not only_GW)):
                    if False:
                        #distance_samples =CF.sample(0.1,5,signal,bound=False,Nsample=Nsamp_GW)
                        #distance_samples_GW=distance_samples[0,:]*self.binaries_dict['det_luminosity_distance'][i]
                        #cosi_samples_GW=distance_samples[1,:]
                        #kde_GW=gaussian_kde(np.vstack([distance_samples_GW,cosi_samples_GW]))
                        kde_GW=CF.gkde(0.2,4,signal,Nsample=Nsamp_GW)
                        cosi_samples_EM = np.cos(np.random.randn(Nsamp_EM)*self.sigma_iota_EM+self.binaries_dict['det_iota_EM'][i])
                        distance_samples_EM = np.random.randn(Nsamp_EM)*self.sigma_d_EM+self.binaries_dict['det_luminosity_distance_EM'][i]
                        weights=kde_GW(np.vstack([distance_samples_EM/self.binaries_dict['det_luminosity_distance'][i] \
                        ,cosi_samples_EM]))
                        weights[np.isnan(weights)]=0
                        weights/=weights.sum()
                        kde_EM=gaussian_kde(np.vstack([distance_samples_EM,cosi_samples_EM]),weights=weights)
                        distance_samples=kde_EM.resample(Nsamp_comb)[0,:]

                    if True:
                        distance_samples =CF.sample(0.2,4,signal,bound=False,Nsample=Nsamp_GW,cosimin=-1.2,cosimax=1.2)
                        distance_samples_GW=distance_samples[0,:]*self.binaries_dict['det_luminosity_distance'][i]
                        cosiota_samples_GW=distance_samples[1,:]

                        iota_samples_EM = np.random.randn(Nsamp_EM)*self.sigma_iota_EM+self.binaries_dict['det_iota_EM'][i]
                        kde_EM=gaussian_kde(iota_samples_EM,weights=np.abs(np.sin(iota_samples_EM)))

                        weights=kde_EM(np.arccos(cosiota_samples_GW))
                        print('Fraction sick cosiota {:.2f}'.format(len(np.where(np.isnan(weights) | np.isinf(weights))[0])/len(weights)))
                        weights[np.isnan(weights) | np.isinf(weights)]=0
                        weights/=weights.sum()


                        kde_comb=gaussian_kde(np.vstack([distance_samples_GW,cosiota_samples_GW]),weights=weights)
                        ss=kde_comb.resample(Nsamp_comb)
                        distance_samples=ss[0,:]
                        cosiota_samples=ss[1,:]
                else:
                    distance_samples =CF.sample(0.2,4,signal,bound=True,Nsample=Nsamp_GW)
                    distance_samples_GW=distance_samples[0,:]*self.binaries_dict['det_luminosity_distance'][i]
                    cosiota_samples_GW=distance_samples[1,:]
                    distance_samples=distance_samples_GW
                    cosiota_samples=cosiota_samples_GW

            else:

                if (self.binaries_dict['GW_detected'][i] & self.binaries_dict['EM_detected'][i] &
                self.binaries_dict['sky_area_detected'][i] & (not only_GW)):
                    if False:
                        #distance_samples =CF.sample(0.1,5,signal,bound=False,Nsample=Nsamp_GW)
                        #distance_samples_GW=distance_samples[0,:]*self.binaries_dict['det_luminosity_distance'][i]
                        #cosi_samples_GW=distance_samples[1,:]
                        #kde_GW=gaussian_kde(np.vstack([distance_samples_GW,cosi_samples_GW]))
                        kde_GW=CF.gkde(0.2,4,signal,Nsample=Nsamp_GW)
                        cosi_samples_EM = np.cos(np.random.randn(Nsamp_EM)*self.sigma_iota_EM+self.binaries_dict['det_iota_EM'][i])
                        distance_samples_EM = np.random.randn(Nsamp_EM)*self.sigma_d_EM+self.binaries_dict['det_luminosity_distance_EM'][i]
                        weights=kde_GW(np.vstack([distance_samples_EM/self.binaries_dict['det_luminosity_distance'][i]
                        ,cosi_samples_EM]))
                        weights[np.isnan(weights)]=0
                        weights/=weights.sum()
                        kde_EM=gaussian_kde(np.vstack([distance_samples_EM,cosi_samples_EM]),weights=weights)
                        distance_samples=kde_EM.resample(Nsamp_comb)[0,:]

                    if True:
                        distance_samples =CF.sample(0.2,4,signal,bound=False,Nsample=Nsamp_GW,cosimin=-1.2,cosimax=1.2)
                        distance_samples_GW=distance_samples[0,:]*self.binaries_dict['det_luminosity_distance'][i]
                        cosiota_samples_GW=distance_samples[1,:]

                        iota_samples_EM = np.random.randn(Nsamp_EM)*self.sigma_iota_EM+self.binaries_dict['det_iota_EM'][i]
                        kde_EM=gaussian_kde(iota_samples_EM,weights=np.abs(np.sin(iota_samples_EM)))

                        weights=kde_EM(np.arccos(cosiota_samples_GW))
                        print('Fraction sick cosiota {:.2f}'.format(len(np.where(np.isnan(weights) | np.isinf(weights))[0])/len(weights)))
                        weights[np.isnan(weights) | np.isinf(weights)]=0
                        weights/=weights.sum()

                        kde_comb=gaussian_kde(np.vstack([distance_samples_GW,cosiota_samples_GW]),weights=weights)
                        ss=kde_comb.resample(Nsamp_comb)
                        distance_samples=ss[0,:]
                        cosiota_samples=ss[1,:]
                else:
                    distance_samples =CF.sample(0.2,4,signal,bound=True,Nsample=Nsamp_GW)
                    distance_samples_GW=distance_samples[0,:]*self.binaries_dict['det_luminosity_distance'][i]
                    cosiota_samples_GW=distance_samples[1,:]
                    distance_samples=distance_samples_GW
                    cosiota_samples=cosiota_samples_GW


            kde_distance=gaussian_kde(distance_samples)
            dif=np.linspace(np.min(distance_samples),np.max(distance_samples),10000)
            kde_distance=interp1d(dif,kde_distance(dif),bounds_error=False,fill_value=np.min(kde_distance(dif)))

            pdf_single=np.zeros_like(H0_posterior)

            for j in range(len(H0_posterior)):

                eq_redshift_to_dgw= lambda z: GW_luminosity_distance(z,H0=H0_posterior[j],mode=self.distance_gen)

                if self.redshift_gen == 'uniform_comoving_volume':
                    redshift_pop_prior = lambda z: np.exp(uniform_comoving_volume(z,H0=H0_posterior[j]))
                elif self.redshift_gen == 'z_square':
                    redshift_pop_prior = lambda z: np.exp(z_square(z))
                elif self.redshift_gen == 'z_uniform':
                    redshift_pop_prior = lambda z: np.exp(z_uniform(z))
                if self.redshift_gen == 'uniform_comoving_volume_first_order':
                    redshift_pop_prior = lambda z: np.exp(uniform_comoving_volume_first_order(z,H0=H0_posterior[j]))
                else:
                    sys.exit('not yet implemented')



                if self.redshift_gen == 'uniform_comoving_volume':
                    prior=np.array([redshift_pop_prior(zz) for zz in z_samples])
                else:
                    prior=redshift_pop_prior(z_samples)

                if self.distance_gen == 'full':
                    dist_vals=np.array([eq_redshift_to_dgw(zz) for zz in z_samples])
                else:
                    dist_vals=eq_redshift_to_dgw(z_samples)

                pdf_single[j] = np.log(np.sum(kde_distance(dist_vals)*prior))


            if isinstance(injections, str):
                if (injections=='GW_EM') & (not only_GW):
                    pdf_single-=self.log_sel_function_interpolation(H0_posterior)
                elif (injections=='GW_EM') & (only_GW):
                    pdf_single-=self.log_sel_function_interpolation_GW(H0_posterior)
                elif (injections=='GW_EM_fow') & (not only_GW):
                    pdf_single-=self.log_sel_function_interpolation_area(H0_posterior)
                elif (injections=='GW_EM_fow') & (only_GW):
                    pdf_single-=self.log_sel_function_interpolation_GW(H0_posterior)
            else:
                if only_GW:
                    pdf_single-=self.log_sel_function_interpolation_GW(H0_posterior)
                elif (self.binaries_dict['GW_detected'][i] &
                self.binaries_dict['EM_detected'][i] &
                self.binaries_dict['sky_area_detected'][i]):
                    pdf_single-=self.log_sel_function_interpolation_area(H0_posterior)
                elif (self.binaries_dict['GW_detected'][i] &
                self.binaries_dict['EM_detected'][i]):
                    pdf_single-=self.log_sel_function_interpolation(H0_posterior)
                else:
                    pdf_single-=self.log_sel_function_interpolation_GW(H0_posterior)

            pdf_single-=np.log(np.sum(np.exp(pdf_single)*dH0))
            pdf_single_out.append(pdf_single)
            pdf_combined+=pdf_single
            pdf_combined-=np.log(np.sum(np.exp(pdf_combined)*dH0))
            i_inj +=1

            if i_inj==num_inj:
                break

        return pdf_single_out,pdf_combined

    def save_pickle(self,filename):

        copy_to_save=copy.deepcopy(self)
        delattr(copy_to_save, 'detector_list')
        file = open(filename, 'wb')
        pickle.dump(copy_to_save,file)
        file.close()

    def plot_likelihood(self,num_inj,
    Nsamp_GW=5000,Nsamp_EM=5000,Nsamp_comb=10000,Nsamp_z=100):

        fig, ax = plt.subplots(2,2,figsize=(10,10))

        c=299792.458 # Speed of light in km/s
        z_samples = self.binaries_dict['z_detected'][num_inj]+np.random.randn(Nsamp_z)*(self.sigma_z/c)


        CF = log_CF_likelihood(self.detector_list)
        signal={'tgps': np.array([self.binaries_dict['tgps'][num_inj]]),
            'dec': np.array([self.binaries_dict['dec'][num_inj]]),
            'RA': np.array([self.binaries_dict['RA'][num_inj]]),
            'psi': np.array([self.binaries_dict['psi'][num_inj]]),
            'iota': np.array([self.binaries_dict['det_iota'][num_inj]]),
            'SNR_detected': np.array([self.binaries_dict['SNR_detected'][num_inj]])}

        distance_samples =CF.sample(0.2,2,signal,bound=True,Nsample=Nsamp_GW)
        distance_samples_GW=distance_samples[0,:]*self.binaries_dict['det_luminosity_distance'][num_inj]
        cosiota_samples_GW=distance_samples[1,:]

        iota_samples_EM = np.random.randn(Nsamp_EM)*self.sigma_iota_EM+self.binaries_dict['det_iota_EM'][num_inj]
        kde_EM=gaussian_kde(iota_samples_EM,weights=np.abs(np.sin(iota_samples_EM)))

        weights=kde_EM(np.arccos(cosiota_samples_GW))
        weights[np.isnan(weights)]=0
        weights/=weights.sum()

        kde_comb=gaussian_kde(np.vstack([distance_samples_GW,cosiota_samples_GW]),weights=weights)
        samples_comb=kde_comb.resample(Nsamp_comb)

        ax[0,0].hist(z_samples,bins='auto',alpha=0.5)
        ax[0,0].axvline(self.binaries_dict['z'][num_inj]
        ,ymin=0,ymax=ax[0,0].get_ylim()[1],color='r',linestyle='--',label='Truth')
        ax[0,0].axvline(self.binaries_dict['z_detected'][num_inj]
        ,ymin=0,ymax=ax[0,0].get_ylim()[1],color='k',linestyle='--',label='z-Detected')
        ax[0,0].set_xlabel('z')
        ax[0,0].legend()

        ax[0,1].hist(np.cos(iota_samples_EM),bins='auto',alpha=0.5)
        ax[0,1].axvline(np.cos(self.binaries_dict['iota'][num_inj])
        ,ymin=0,ymax=ax[0,1].get_ylim()[1],color='r',linestyle='--',label='Truth')
        ax[0,1].axvline(np.cos(self.binaries_dict['det_iota_EM'][num_inj])
        ,ymin=0,ymax=ax[0,1].get_ylim()[1],color='g',linestyle='--',label='EM-Detected')
        ax[0,1].set_xlabel(r'$\cos \iota$')
        ax[0,1].legend()

        #sns.kdeplot(distance_samples_GW,cosiota_samples_GW,ax=ax[1,0])
        ax[1,0].axhline(np.cos(self.binaries_dict['det_iota_EM'][num_inj]),xmin=ax[1,0].get_xlim()[0],
        xmax=ax[1,0].get_xlim()[1],color='g',label='EM-Detected',linestyle='--')
        plot_2d_kde(ax[1,0],distance_samples_GW,cosiota_samples_GW,
             cmap=cm.Blues, shade=True, levels=[0.683, 0.954],alpha=0.3)
        plot_2d_kde(ax[1,0],distance_samples_GW,cosiota_samples_GW,
             cmap=None, shade=False, levels=[0.683, 0.954],linestyles='solid',colors='b',linewidths=1)

        ax[1,0].set_xlabel(r'$d_L$ [Mpc]')
        ax[1,0].set_ylabel(r'$\cos \iota$')
        ax[1,0].scatter(self.binaries_dict['luminosity_distance'][num_inj],
        np.cos(self.binaries_dict['iota'][num_inj]),c='r',label='Truth')
        ax[1,0].scatter(self.binaries_dict['det_luminosity_distance'][num_inj],
        np.cos(self.binaries_dict['det_iota'][num_inj]),c='orange',label='GW-Detected')
        ax[1,0].legend()

        #sns.kdeplot(samples_comb[0,:],samples_comb[1,:],ax=ax[1,1])
        ax[1,1].axhline(np.cos(self.binaries_dict['det_iota_EM'][num_inj]),xmin=ax[1,1].get_xlim()[0],
        xmax=ax[1,1].get_xlim()[1],color='g',label='EM-Detected',linestyle='--')
        plot_2d_kde(ax[1,1],samples_comb[0,:],samples_comb[1,:],
             cmap=cm.Blues, shade=True, levels=[0.683, 0.954],alpha=0.3)
        plot_2d_kde(ax[1,1],samples_comb[0,:],samples_comb[1,:],
             cmap=None, shade=False, levels=[0.683, 0.954],linestyles='solid',colors='b',linewidths=1)
        ax[1,1].set_xlabel(r'$d_L$ [Mpc]')
        ax[1,1].set_ylabel(r'$\cos \iota$')
        ax[1,1].scatter(self.binaries_dict['luminosity_distance'][num_inj],
        np.cos(self.binaries_dict['iota'][num_inj]),c='r',label='Truth')
        ax[1,1].scatter(self.binaries_dict['det_luminosity_distance'][num_inj],
        np.cos(self.binaries_dict['det_iota'][num_inj]),c='orange',label='GW-Detected')
        ax[1,1].legend()




def load_ana(filename,detector_list):

    file = open(filename, 'rb')
    ciao = pickle.load(file)
    ciao.detector_list = detector_list
    file.close()
    return ciao
