import os
import shutil
import numpy as np
import  matplotlib.pyplot as plt
import time
from scipy.stats import gaussian_kde
import bilby

from ..detector.detection_probability import PdetGW_compute
from ..cosmology_functions.astrofunction import dephasing_conversion_tauk,time_delay_GW_EM
from ..cosmology_functions.astrofunction import GW_luminosity_distance,Cm_GW_luminosity_distance
from ..statistic.likelihoods import log_gaussian_kde_likelihood
from ..statistic.likelihoods import testingGRlikelihood

class testingGRcosmo(object):
    """ A Class for performing a Bayesian Analysis in Effective Field theory with GWs

    Parameters
    ----------
    dict_samples: dictionary
        Dict where you have stored the luminosity distance samples, GW-EM delay samples,
        tauk samples and redfhist samples. The dictionary should have the following structure
        *GW_data* with sub fields *distance* (Luminosity distance in Mpc) and *phase* (Dephasing in tauk)
        *GW_EM_delay* (GW-EM samples in seconds). *redshift* with subfields *z_obs* (Observed redshift)
        and *z_pec* (Peculiar redshift)
    lambda_am: lambda equation
        Lambda Eq. for alpha_M (z). *z* must be the first variable
    lambda_at: lambda equation
        Lambda Eq. for alpha_tk (z). *z* must be the first variable
    zmax: float
        Maximum redshift for computing selection effects. Should be sufficiently high (e.g. 1.0)
    horizon: float, optional
        Horizon in Mpc for the GW detection of the detector network. If given assumes the
        detection probability as a box. If given you must not give detector_anal
    detectors_anal: list of detector objects, optional
        List of detector objects containing the GW detectors. It will compute Pdet with
        a monte carlo
    SNR_thr: float, optional
        SNR for the detection. To be given only if given **detectors_anal**
    mass_generation: str, optional
        Mass generation for the binary. See functions available in the package population
        To be given only if given **detectors_anal**
    alpha_mass: float, optional
        Power low coefficient for the mass generation of BBHs.
        To be given only if given **detectors_anal**
    Nsample: int, optional
        Number of binaries to simulate for computing Pdet.
        To be given only if given **detectors_anal**
    k_model: float
        Dispersion relation coefficient. E.g. for graviton -2.
    outdir: str
        Outdirt where to save the results
    """
    def __init__(self, dict_samples,
    lambda_am, lambda_at, zmax, horizon=None,detectors_anal=None,
    SNR_thr=None, k_model=0.,outdir=time.strftime("%Y_%m_%d_%H_%M_%S"),
    mass_generation='BNS-gaussian', alpha_mass=1.6, Nsample=5000):

        self.zmax=zmax
        self.k_model=k_model # k of the alpha_T freq. dependence
        self.outdir=outdir

        if os.path.exists(outdir): # Check if folder is present and delete it
            shutil.rmtree(outdir)
        os.makedirs(outdir)
        self.eq_am=lambda_am
        self.eq_at=lambda_at
        self.k_model = k_model

        if (horizon is None) and (detectors_anal is None):
            sys.exit('You should give an Horizon or a detectors list, not both two')
        if (not (horizon is None)) and not ((detectors_anal is None)):
            sys.exit('You should give an Horizon or a detectors list, just one')

        if detectors_anal is None:
            Pd,dcrit=PdetGW_compute(fast_method=True,horizon=horizon)
            self.PdetGW=Pd
            self.dcrit_Pdet=dcrit
        else:
            Pd,dcrit=PdetGW_compute(fast_method=False,detectors = detectors_anal, SNR_net_thr=SNR_thr,
mass_generation=mass_generation, alpha=alpha_mass, Nsample=Nsample)
            self.PdetGW=Pd
            self.dcrit_Pdet=dcrit
            self.mass_generation=mass_generation
            self.alpha_mass=alpha_mass
            self.Nsample=Nsample
            self.SNR_net_thr=SNR_thr
            #self.detectors_list=detectors_anal

        d_plot=np.linspace(0.5,dcrit,1000)
        plt.figure()
        plt.plot(d_plot,Pd(d_plot))
        plt.xlabel('Distance [Mpc]')
        plt.ylabel(r'$P_{\rm det GW}$')
        plt.title('GW detection probability (single source)')
        plt.savefig(self.outdir+'/detection_probability.png')

        self._pepare_samples_probability(dict_samples)

    def _pepare_samples_probability(self,dict_samples):
        """ This method fit and plot in the output directory the gaussian kernels
        to the posterior sample provided

        Parameters
        ----------
        dict_samples: dictionary
            Dict where you have stored the luminosity distance samples, GW-EM delay samples,
            tauk samples and redfhist samples. The dictionary should have the following structure
            *GW_data* with sub fields *distance* (Luminosity distance in Mpc) and *phase* (Dephasing in tauk)
            *GW_EM_delay* (GW-EM samples in seconds). *redshift* with subfields *z_obs* (Observed redshift)
            and *z_pec* (Peculiar redshift)
        """

        dict_kernel={}

        for key_dict in dict_samples.keys():

            if key_dict == 'GW_data':
                samples_given = dict_samples[key_dict].keys()
                if len(dict_samples[key_dict]) == 2:

                    plt_arr_distance=np.linspace(np.min(dict_samples['GW_data']['distance']),
                                         np.max(dict_samples['GW_data']['distance']), 100)
                    plt_arr_tauk=np.linspace(np.min(dict_samples['GW_data']['phase']),
                                         np.max(dict_samples['GW_data']['phase']), 100)

                    to_fit=np.vstack([dict_samples['GW_data']['distance'],dict_samples['GW_data']['phase']])
                    tmp_kernel_GW_data = gaussian_kde(to_fit)
                    dist_vals, phase_vals = np.meshgrid(plt_arr_distance, plt_arr_tauk)
                    positions = np.vstack([dist_vals.ravel(), phase_vals.ravel()])
                    pdf_values = tmp_kernel_GW_data(positions)
                    dict_kernel['GW_data']=tmp_kernel_GW_data
                    plt.figure()
                    plt.contourf(dist_vals,phase_vals,np.reshape(pdf_values,dist_vals.shape))
                    plt.colorbar()
                    plt.xlim([plt_arr_distance[0],plt_arr_distance[-1]])
                    plt.ylim([plt_arr_tauk[0],plt_arr_tauk[-1]])
                    plt.xlabel('Distance [Mpc]')
                    plt.ylabel('tauk')
                    plt.title('Posterior GW data')
                    plt.savefig(self.outdir+'/distance_tauk_prob.png')

                else:
                    passed_sample = list(dict_samples['GW_data'].keys())[0] # Read distance or phase
                    tmp_kernel_gws=gaussian_kde(dict_samples['GW_data'][passed_sample])
                    dl_vals=np.linspace(np.min(dict_samples['GW_data'][passed_sample]),
                                     np.max(dict_samples['GW_data'][passed_sample]),500)

                    dl_pdf=tmp_kernel_gws(dl_vals)

                    dd=np.mean(np.diff(dl_vals))
                    dict_kernel['GW_data']={}
                    dict_kernel['GW_data'][passed_sample]=tmp_kernel_gws
                    plt.figure()
                    plt.plot(dl_vals,dl_pdf)
                    plt.xlabel(passed_sample)
                    plt.ylabel('Posterior')
                    plt.savefig(self.outdir+'/GW_data_prob.png')

            elif key_dict == 'GW_EM_delay':

                tmp_kernel_t=gaussian_kde(dict_samples['GW_EM_delay'])
                dl_vals=np.linspace(np.min(dict_samples['GW_EM_delay']),
                                 np.max(dict_samples['GW_EM_delay']),500)
                dx = np.mean(np.diff(dl_vals))

                dl_pdf=tmp_kernel_t(dl_vals)*dx # This is a probability directly

                dict_kernel['GW_EM_delay']=tmp_kernel_t
                plt.figure()
                plt.plot(dl_vals,dl_pdf)
                plt.xlabel('GW-EM time delay [s]')
                plt.ylabel('Probability')
                plt.savefig(self.outdir+'/initial_GW_EM_time_delay_posterior.png')

            elif  key_dict == 'redshift':

                samples_given = dict_samples[key_dict].keys()
                if len(dict_samples[key_dict]) == 2:

                    plt_arr_distance=np.linspace(np.min(dict_samples['redshift']['zobs']),
                                         np.max(dict_samples['redshift']['zobs']), 100)
                    plt_arr_tauk=np.linspace(np.min(dict_samples['redshift']['zpec']),
                                         np.max(dict_samples['redshift']['zpec']), 100)

                    to_fit=np.vstack([dict_samples['redshift']['zobs'],dict_samples['redshift']['zpec']])
                    dz1=np.mean(np.diff(plt_arr_distance))
                    dz2=np.mean(np.diff(plt_arr_tauk))
                    tmp_kernel_z = gaussian_kde(to_fit)
                    dist_vals, phase_vals = np.meshgrid(plt_arr_distance, plt_arr_tauk)
                    positions = np.vstack([dist_vals.ravel(), phase_vals.ravel()])
                    pdf_values = tmp_kernel_z(positions)*dz1*dz2 # This is probability
                    dict_kernel['redshift']=tmp_kernel_z
                    plt.figure()
                    plt.contourf(dist_vals,phase_vals,np.reshape(pdf_values,dist_vals.shape))
                    plt.colorbar()
                    plt.xlim([plt_arr_distance[0],plt_arr_distance[-1]])
                    plt.ylim([plt_arr_tauk[0],plt_arr_tauk[-1]])
                    plt.xlabel('zobs')
                    plt.ylabel('zpec')
                    plt.title('Posterior')
                    plt.savefig(self.outdir+'/z_obs_zpec_prob.png')

                else:
                    passed_sample = list(dict_samples['redshift'].keys())[0] # Read distance or phase
                    tmp_kernel_zs=gaussian_kde(dict_samples['redshift'][passed_sample])
                    dl_vals=np.linspace(np.min(dict_samples['redshift'][passed_sample]),
                                     np.max(dict_samples['redshift'][passed_sample]),500)
                    dl_pdf=tmp_kernel_zs(dl_vals)

                    dzs=np.mean(np.diff(dl_vals))
                    dl_pdf=dzs*dl_pdf

                    dict_kernel['redshift']={}
                    dict_kernel['redshift'][passed_sample]=tmp_kernel_zs

                    plt.figure()
                    plt.plot(dl_vals,dl_pdf)
                    plt.xlabel(passed_sample)
                    plt.ylabel('Posterior')
                    plt.savefig(self.outdir+'/redshift_prob.png')


            else:
                raise 'Type of samples not recognized'

        self.probabilities_from_samples=dict_kernel

    def likelihood_full_parameters(self, H0, Om0, zcosmo, at, am, tau, freq_merger, zpec = None,distance_model='GW_luminosity_distance'):
        """ This method returns the log likelihood of the testing GR model. This is given by
        P(x_GW|dpsi, distance) P(x_em|z_obs, z_pec) P(x_GRB | Delta t). It also removes the d^2 prior.

        Parameters
        ----------
        H0: float
            Hubble constant in km/Mpc s
        Om0: float
            Dark matter density today
        zcosmo: float
            cosmological redshift
        at: list
            Coefficients of the lambda Eq. for alpha_T
        am: list
            Coefficients of the lambda Eq. for alpha_M
        tau: float
            initial prompt delay between GW-EM
        freq_merger: float
            frequency of the merger or ISCO. I.e. the frequency of the GW when used to
            compute the GW-EM time delay.
        z_pec: float, optional
            peculiar redshift.

        Returns
        -------
        log likelihood: float
            The log likelihood
        """

        prob_kernels = self.probabilities_from_samples
        prob_GW_data = 0.
        prob_redshift = 0.
        prob_GW_EM_delay = 0.

        for data_kerneles in prob_kernels.keys():

            if data_kerneles == 'GW_EM_delay':
                time_template, _ = time_delay_GW_EM(zcosmo,k=self.k_model,
                freq_merger=freq_merger,alpha_Tk=at, lambda_alpha_T=self.eq_at ,
                initial_prompt_delay=tau,Om0=Om0,H0=H0)
                template = np.array([time_template])
                kde_fit = prob_kernels['GW_EM_delay']
                prob_GW_EM_delay = log_gaussian_kde_likelihood(kde_fit,template)

            if data_kerneles == 'redshift':
                if zpec is None:
                    template = np.array([zcosmo])
                    kde_fit = prob_kernels['redshift']['zobs']
                    prob_redshift = log_gaussian_kde_likelihood(kde_fit,template)

                if ((zcosmo is not None) and (zpec is not None)):
                    obs_red=(1.+zcosmo)*(1.+zpec)-1.
                    template = np.array([obs_red, zpec])
                    kde_fit = prob_kernels['redshift']
                    prob_redshift = log_gaussian_kde_likelihood(kde_fit,template)

            if data_kerneles == 'GW_data':

                if distance_model=='GW_luminosity_distance':
                    d_template= GW_luminosity_distance(zcosmo,alpha_M=am,
                    lambda_alpha_M=self.eq_am,Om0=Om0,H0=H0)
                elif distance_model=='Cm_GW_luminosity_distance':
                    d_template= Cm_GW_luminosity_distance(zcosmo,cm=am[0],Om0=Om0,H0=H0)

                _, tauk_template = time_delay_GW_EM(zcosmo,k=self.k_model,
                freq_merger=freq_merger,alpha_Tk=at, lambda_alpha_T=self.eq_at ,
                initial_prompt_delay=tau,Om0=Om0,H0=H0)

                try:
                    template = np.array([d_template, tauk_template])
                    kde_fit = prob_kernels['GW_data']
                    #Remove the d^2 prior below
                    prob_GW_data = log_gaussian_kde_likelihood(kde_fit,template)-2*np.log(d_template)
                except:
                    param=list(prob_kernels['GW_data'].keys())[0] # Read the parameter we are running
                    if param == 'distance':
                        template = np.array([d_template])
                        kde_fit = prob_kernels['GW_data']['distance']
                        #Remove the d^2 prior below
                        prob_GW_data = log_gaussian_kde_likelihood(kde_fit,template)-2*np.log(d_template)
                    if param == 'phase':
                        template = np.array([tauk_template])
                        kde_fit = prob_kernels['GW_data']['phase']
                        #Remove the d^2 prior below
                        prob_GW_data = log_gaussian_kde_likelihood(kde_fit,template)-2*np.log(d_template)


        result = prob_GW_data+prob_redshift+prob_GW_EM_delay

        return result

    def run_bilby(self,priors,likelihood=testingGRlikelihood,red_prior='uniform_comoving_volume',
    pop_prior='uniform_comoving_volume',distance_model='GW_luminosity_distance',**kwargs):
        """This method run bilby using dynesty.

        Parameters
        ----------
        priors: dict
            Dictionary of bilby priors of the parameters. These are *H0*, *Om0*, *alpha_M*, *alpha_T*
            *freq_merger*, *z cosmo*, *zpec*.
        red_prior: str
            Type of prior to use as prior for the redshift measurement. These are *uniform_comoving_volume*
            *z_square*, *z_uniform*, *uniform_comoving_volume_time_star_rate*
        pop_prior: str
            Type of prior to use as prior for the population and selection effects. These are *uniform_comoving_volume*
            *z_square*, *z_uniform*, *uniform_comoving_volume_time_star_rate*

        Returns
        -------
        results: class
            Bilby result class for the sampler.
        """

        for ss in priors.keys():
            if (not priors[ss].name == ss):
                print('There is an error with the names in the priors for, pls use same name'+ss+'-'+'priors[ss].name')
                sys.exit('Error on the names of the priors')

        result=bilby.run_sampler(likelihood=testingGRlikelihood(self, red_prior=red_prior,
        pop_prior=pop_prior,distance_model=distance_model), priors=priors,outdir=self.outdir
        , **kwargs)
        return result
