import numpy as np
from scipy.integrate import quad
from scipy.stats import gaussian_kde
from sys import exit


def E_z(z,Om0=0.3):
    """
    Computes the FLRW parameter E(z)

    Parameters
    ----------
    z: float
        The cosmological redshift
    Om0: float, optional
        Dark matter density today

    Returns
    -------
    float
        GW luminosity distance in Mpc
    """
    return (Om0*((1+z)**3.)+1-Om0)**(0.5)

def Cm_GW_luminosity_distance(z,cm=0,Om0=0.3,H0=67.7329):
    """
    Computes the GW luminosity distance as in Eq. 20 of (https://arxiv.org/pdf/1901.03321.pdf)

    Parameters
    ----------
    z: float
        The cosmological redshift
    cm: float
        modified friction term.
    Om0: float
        Matter density today
    H0: float
        Hubble parameter today in km/Mpc s

    Returns
    -------
    float
        GW luminosity distance in Mpc
    """

    c=299792458.0 # Speed of light in m/s
    H0=H0*3.24078e-20 #Conversion to s-1

    FRLW_integrand = lambda z_lam : (Om0*((1+z_lam)**3.)+1-Om0)**(-0.5)
    d_L_LCDM=quad(FRLW_integrand,0,z)[0]*c*(1+z)/H0 # Computes luminosity distance in FLRW
    d_L_LCDM = c*z/H0

    dlgw = d_L_LCDM*(1.+0.5*cm*z+(cm/8.)*(cm-2.-6.*Om0)*z**2)

    return dlgw*3.24078e-23


def GW_luminosity_distance(z,alpha_M=None,lambda_alpha_M=None,Om0=0.3,H0=67.7329,mode='full'):
    """
    Computes the GW luminosity distance in modified gravity

    Parameters
    ----------
    z: float
        The cosmological redshift
    alpha_M: list or numpy array
        The coefficients of the deviation alpha_M. The order of coefficient
        should be consistent with lambda_alpha_M.
    lambda_alpha_M: lambda Eq.
        The Eq. for alpha_M. Redhshift should be the first parameter must be
        the cosmological redshift.
    Om0: float
        Matter density today
    H0: float
        Hubble parameter today in km/Mpc s

    Returns
    -------
    float
        GW luminosity distance in Mpc
    """

    if mode =='full':

        if (not (alpha_M is None)) and (not (lambda_alpha_M is None)):
            if lambda_alpha_M.__code__.co_varnames[0] != 'z' :
                exit('Redhift must be first var')
            c=299792458.0 # Speed of light in m/s
            H0=H0*3.24078e-20 #Conversion to s-1

            FRLW_integrand = lambda z_lam : (Om0*((1+z_lam)**3.)+1-Om0)**(-0.5)
            d_L_LCDM=quad(FRLW_integrand,0,z)[0]*c*(1+z)/H0
            mod_factor_int= lambda z_val: lambda_alpha_M(z_val,*alpha_M)/(1.+z_val)
            mod_factor=np.exp(quad(mod_factor_int,0.,z)[0])

            return d_L_LCDM*mod_factor*3.24078e-23 # Distance in MPCs

        if (alpha_M is None) and (lambda_alpha_M is None):
            c=299792458.0 # Speed of light in m/s
            H0=H0*3.24078e-20 #Conversion to s-1

            FRLW_integrand = lambda z_lam : (Om0*((1+z_lam)**3.)+1-Om0)**(-0.5)
            d_L_LCDM=quad(FRLW_integrand,0,z)[0]*c*(1+z)/H0
            return d_L_LCDM*3.24078e-23 # Distance in MPCs

    elif mode =='linear':
        c=299792.458 # Speed of light in km/s
        return c*z/H0

    elif mode =='first_order':
        c=299792.458 # Speed of light in km/s
        q0 = 0.5*Om0-(1-Om0)
        return (c*z/H0)*(1+0.5*(1-q0)*z)


def time_delay_GW_EM(z,k= None,freq_merger= None,alpha_Tk = None, lambda_alpha_T = None,initial_prompt_delay=0.,Om0=0.3,H0=67.7329):
    """ Computes the GW-EM time delay and the coefficient tau_k

    Parameters
    ----------
    z: float
        The cosmological redshift
    k: float
        power of the frequency dependence for dispersion relation  (e.g graviton =2)
    freq_merger: float
        Frequency of the merger or ISCO in Hz
    alpha_Tk: list or numpy array
        The coefficients of the deviation alpha_T. The order of coefficient
        should be consistent with lambda_alpha_T.
    lambda_alpha_T: lambda Eq.
        The Eq. for alpha_T with respect to redshift. Redhshift should be the first parameter must be
        the cosmological redshift.
    initial_prompt_delay: float
        Initial prompt delay in seconds betweem GW-EM
    Om0: float
        Matter density today
    H0: float
        Hubble parameter today in km/Mpc s

    Returns
    -------
    float
        GW-EM time delay in s
    float
        tau_k parameter
    """
    if (not (alpha_Tk is None)) and (not (lambda_alpha_T is None)):

        if lambda_alpha_T.__code__.co_varnames[0] != 'z' :
            exit('Redhift must be first var')

        H0=H0*3.24078e-20 #Conversion to s-1
        int_function= lambda z_val: (np.power(1.0+z_val,k)/E_z(z_val, Om0 = Om0))*(lambda_alpha_T(z_val,*alpha_Tk)/H0)
        t_k=quad(int_function,0,z)[0]

        return (1.+z)*initial_prompt_delay+0.5*(freq_merger**k)*t_k, t_k
    else:
        return (1.+z)*initial_prompt_delay,0.

def compute_condition_alpha(z, Om0=0.3):
    """ Computes a number X  in such a way that if alpha_M(z) > X, then the GW
    luminosity distance is crescent in (z)

    Parameters
    ----------
    z : float
        cosmological redshift
    Om0 : float
        Matter density today

    Returns
    -------
    float
        Maximum value of alpha at given redshift
    """

    if z<1e-6:
        exit('Warning, use a redshift > than 1e-6')

    func_to_int= lambda z_lam: 1./E_z(z_lam,Om0=Om0)
    condition_alpha_lam= lambda z_lam : (((1+z_lam)/E_z(z_lam))/quad(func_to_int,0,z_lam)[0])+1

    return -1.*condition_alpha_lam(z)

def compute_chi(mass_1, mass_2, spin_1, spin_2):
    """ Computes
    Compute chi_p for CBC.
    mass_1: In solar masses
    mass_2: In solar masses
    spin_1: In solar masses
    spin_2: In solar masses
    """
    totalMass = mass_1+mass_2
    eta = mass_1*mass_2/totalMass**2.
    delta = np.sqrt(1.-4.*eta)
    chi = 0.5*(spin_1*(1.+delta) + spin_2*(1.-delta))
    return chi



def PN_coefficient_TaylorF2_reducedSpins(mass_1,mass_2,spin_1,spin_2,order,log_term,tgr=False):
    """ Return the PN coefficient (https://arxiv.org/pdf/0907.0700.pdf) Eq.3.18.
    Check (https://lscsoft.docs.ligo.org/lalsuite/lalsimulation/_l_a_l_sim_inspiral_taylor_f2_reduced_spin_tidal_8c_source.html)
    for compatibility
    Note that Tidal components scale as f^((10-5)/3) and f^((12-5)/3) so we do not consider them

    Parameters
    ----------
    mass_1: numpy array
        Mass samples of the first object in solar masses
    mass_2: numpy array
        Mass samples of the second object in solar masses
    spin_1: numpy array
        spin samples of the first object geometric units
    spin_2: numpy array
        spin samples of the second object geometric units
    order: int
        Order of the 2*PN expnasion you are feeding. From 0 to 7 included.
    log_temr: boolean
        Do you want the log term of the PN expansion? Only present for order 5,6
    samples_dephasing: numpy array
        samples of the parametrized PN
    tgr: Bool (optional)
        When True, return the alpha PN=1 equal to one

    Returns
    -------
    numpy arrays
        psi0*alpha
    """

    mass_1*=4.925e-6
    mass_2*=4.925e-6
    m=mass_1+mass_2
    eta = mass_1 * mass_2 / m**2.    #Symmetric mass ratio
    #print(mass_1,mass_2,eta)
    chi=compute_chi(mass_1, mass_2, spin_1, spin_2)
    #psi0=(3./(256*eta))*np.power(2*np.pi*m,(order-5.)/3)
    psi3S = 113.*chi/3.
    psi4S = 63845.*(-81. + 4.*eta)*chi*chi/(8.*np.power(-113. + 76.*eta, 2.))
    psi5S = -565.*(-146597. + 135856.*eta + 17136.*eta*eta)*chi/(2268.*(-113. + 76.*eta))

    psi0=(3./(128*eta))*np.power(np.pi*m,(order-5.)/3)


    if log_term:
        if order == 7:
            alfa = 0
        elif order == 6:
            alfa = -6848./63.
        elif order == 5:
            alfa = np.pi*((38645./756)-(65./9)*eta)
        elif order ==4:
            alfa = 0
        elif order == 3:
            alfa = 0
        elif order == 2:
            alfa = 0
        elif order == 1:
            alfa=0
        elif order ==0:
            alfa=0
    else:
        if order == 7:
            alfa= np.pi*(77096675./254016 + 378515.*eta/1512 - 74045.*eta*eta/756)
        elif order == 6:
            alfa = 11583231236531./4694215680. - (640.*np.pi*np.pi)/3. - (6848.*np.euler_gamma)/21. \
                     + (-5162.983708047263 + 2255.*np.pi*np.pi/12.)*eta \
                     + (76055.*eta*eta)/1728. - (127825.*eta*eta*eta)/1296.-6848.*np.log(64*np.pi*m)/63
        elif order == 5:
            alfa = np.pi*(38645./756. - 65.*eta/9.)*(1.+np.log(np.pi*m*6**1.5))+psi5S
        elif order ==4:
            alfa = 10*(3058673./1016064 + 5429.*eta/1008 + 617.*eta*eta/144)+psi4S
        elif order == 3:
            alfa = - 16.*np.pi+psi3S
        elif order == 2:
            alfa = (20./9)*(743./336+11.*eta/4)
        elif order == 1:
            if tgr:
                alfa=1
                # Note that here we set alfa=1 beacuse the definition of dphi_1 is different, \
            #since \psi GR=0 it uses only the normalization psi0
            else:
                alfa=0
        elif order ==0:
            alfa=1

    return psi0*alfa

def LVI_to_tauk(LVI_A_samples,LVI_redshift,LVI_alfa,LVI_H0=69.3,LVI_Om0=0.3):

    """ This function convertes the LVI samples into tauk samples. Do not removes
    Lo prior on LVI

    Parameters
    ----------
    LVI_A_samples: numpy array
        samples on A_alpha from LVI in eV^{2-alfa}/c^{2alfa}
    LVI_redshift_samples: numpy array
        LVI redsfhit samples
    LVI_alfa: float
        float alpha for LVI
    LVI_H0: float
        value of the Hubble constant in LALinference (default 69.3)
    order: int
        Order of the 2*PN expnasion you are feeding. From 0 to 7 included.
    remove_loguni_prior: boolean
        True if you want to remove the log prior
    Returns
    -------
    numpy arrays
        Tauk samples
    """

    kappa=LVI_alfa-2
    LVI_H0=LVI_H0*3.24078e-20 #Conversion to s-1
    hplan=6.583e-16 #in ev*s

    LVI_A_samples=LVI_A_samples*(hplan**(LVI_alfa-2))  #Convert LVI A samples in s^(LVI_alfa-2)

    int_function= lambda z_val: np.power(1.0+z_val,kappa)/E_z(z_val, Om0 = LVI_Om0)

    integral_factor=np.ones_like(LVI_A_samples)

    for i in range(len(LVI_A_samples)):
        integral_factor[i]=quad(int_function,0,LVI_redshift[i])[0]

    if LVI_alfa == 1 :
        tauk_posterior=(LVI_A_samples/LVI_H0)*integral_factor
    else:
        tauk_posterior=((kappa+1)/(LVI_alfa-1))*(LVI_A_samples/LVI_H0)*integral_factor

    return tauk_posterior



def dephasing_conversion_tauk(mass_1,mass_2,spin_1,spin_2,order,log_term,samples_dephasing):
    """ Converts the parametrized PN (https://arxiv.org/pdf/1110.0530.pdf) to tau_k samples.

    Parameters
    ----------
    mass_1: numpy array
        Mass samples of the first object in solar masses
    mass_2: numpy array
        Mass samples of the second object in solar masses
    spin_1: numpy array
        spin samples of the first object geometric units
    spin_2: numpy array
        spin samples of the second object geometric units
    order: int
        Order of the 2*PN expnasion you are feeding. From 0 to 7 included.
    log_temr: boolean
        Do you want the log term of the PN expansion? Only present for order 5,6
    samples_dephasing: numpy array
        samples of the parametrized PN

    Returns
    -------
    numpy arrays
        Tauk samples
    """

    psi0alfa=PN_coefficient_TaylorF2_reducedSpins(mass_1,mass_2,spin_1,spin_2,order,log_term,tgr=True)
    tau_k_order=(order-8)/3.

    if log_term:
        if order != 5:
            exit('We can not conver this PN term to tau_k (see notes)')
        else:
            tau_k_samples=(psi0alfa/np.pi)*samples_dephasing
    else:
        tau_k_samples=((tau_k_order+1.)/np.pi)*psi0alfa*samples_dephasing

    return tau_k_samples
