from pycbc import detector
from scipy.interpolate import interp1d
import os.path
import pycbc
import pycbc.psd
import numpy as np
import scipy



class DetectorClass(object):
    """ A GW detector class used to compute PSD, Antenna patterns etc.

    Parameters
    ----------
    name : str
        A formatted string *H1*, *L1*, *V1*, *K1*, *I1*
    psd : str
        The PSD of the detector in (1/sqrt(Hz)). Can be txt file with two columns (freq and PSD)
        , Lambda function where the input is the frequency (Numpy compatible),
        PyCBC string for PSD (see https://pycbc.org/pycbc/latest/html/pycbc.psd.html)
    cal : error lambda int_function (optional)
        Function w.r.t frequency of the Calibration error. Must be numpy compatible.

    Attributes
    ----------
    name : str
        Name of the detector
    pycbc_detector : :class:
        Pycbc class of the detector
    psd : str
        PSD chosen
    cal_error : error lambda int_function (optional)
        Calibration error
    """
    def __init__(self,name, psd,cal_error=None):
        self.name = name
        self.pycbc_detector= detector.Detector(name)
        self.psd=psd
        self.cal_error=cal_error
    def compute_antenna_patterns(self,ra,dec,psi,tgps):
        """ This method returns the antenna patterns

        Parameters
        ----------
        ra : numpy array or float
            Right ascension in radians
        dec : numpy or float
            declination in radiants
        psi : numpy array or float
            Polarization angle in radiants
        tgps : numpy array or float
            Time of arrival at the SSB

        Returns
        -------
        Flplus : Numpy array
            Antenna pattern of the plus polarization
        Flcross : Numpy array
            Antenna pattern of the plus polarization
        """
        Fplus, Fcross = self.pycbc_detector.antenna_pattern(right_ascension=ra,
                                                            declination=dec, polarization=psi, t_gps=tgps)
        return Fplus, Fcross
    def compute_PSD(self):
        """ This method returns the interpolation of the PSD in 1/sqrt(Hz)

        Returns
        -------
        psd : Interpolation or lambda Function
            Interpolation or lambda function of psd

        """
        if isinstance(self.psd,str):
            if os.path.isfile(self.psd): # Pass here a file with PSD and freq
                data = np.genfromtxt(self.psd)
                self.freq=data[:,0]
                psd = interp1d(data[:,0],data[:,1],fill_value='extrapolate', bounds_error=False)
                return psd
            else:
                freq=np.logspace(np.log10(20),np.log10(2048),10000)
                self.freq=freq
                sens = pycbc.psd.analytical.from_string(self.psd,int((freq[-1])/0.1)+1,0.1,freq[0])
                sens_return=np.zeros_like(freq)
                for i in range(len(sens_return)):
                    sens_return[i]=sens.at_frequency(freq[i])
                freq=freq[sens_return!=0.]
                sens_return=sens_return[sens_return!=0]
                sens_return=np.sqrt(sens_return)
                psd = interp1d(freq,sens_return,fill_value='extrapolate',bounds_error=False)
                return psd # Pass here the string to pycbc_psd
        else:
            psd = self.psd
            return psd
