import numpy as np
import pandas as pd
from ..statistic.priors import uniform_comoving_volume,uniform_comoving_volume_time_star_rate,z_square,z_uniform
from scipy.interpolate import interp2d
from scipy import stats

def generate_binaries(Nsample,redshift_generator,zmax=0.1,mass_generation='BNS-gaussian',alpha=1.6):
    """ This function generates N binaries and return them as a dict of detections

    Parameters
    ----------
    Nsample: integer
        Number of binaries
    redshift_generator: lambda function
        lambda function in z for the probability of finding a binary at redshift z
    zmax: float
        maximum redshift at which to generate binaries.
    mass_generation: string
        Masses generator for the binaries. Choose BNS-gaussian, BNS-uniform, BBH-powerlaw
    alpha: float
        Power low factor for BBH distributions

    Returns
    -------
    binaries: dictionary
        contains arrays for the different variables.
    """

    if mass_generation == 'BNS-gaussian':
            m1, m2 = BNS_gaussian_distribution(Nsample, mean=1.35, sigma=0.15)
    if mass_generation == 'BNS-uniform':
            m1, m2 = BNS_uniform_distribution(Nsample, mmin=1.2, mmax=1.6)
    if mass_generation == 'BBH-powerlaw':
            m1, m2 = BBH_mass_distribution(Nsample, mmin=5., mmax=50., alpha=self.alpha)


    RAs = np.random.rand(Nsample)*2.0*np.pi # R.A. in radians
    r = np.random.rand(Nsample)
    Decs = np.arcsin(2.0*r - 1.0) # Declination in radians
    q = np.random.rand(Nsample)
    incs = np.arccos(2.0*q - 1.0) # Source inclination in radians
    psis = np.random.rand(Nsample)*2.0*np.pi # Psi in radians
    tgps = 1187008882.43*np.ones_like(psis)

    z_binarie=np.linspace(2e-6,zmax,100000)
    prob_binary=[]
    for z in z_binarie:
        prob_binary.append(redshift_generator(z))

    prob_binary=np.array(prob_binary)
    prob_binary/=prob_binary.sum()

    z_binaries=np.random.choice(z_binarie,Nsample,p=prob_binary)

    binaries_dict={'RA' : RAs, 'dec': Decs, 'iota': incs, 'psi': psis,
                   'mass1': m1, 'mass2': m2, 'tgps': tgps,'z':z_binaries}

    return binaries_dict


def BBH_mass_distribution(N, mmin=5., mmax=40., alpha=1.6):
    """
    Returns p(m1,m2)
    The prior on the mass distribution that follows a power
    law for BBHs.

    Parameters
    ----------
    N : integer
        Number of masses sampled
    mmin : float
        minimum mass
    mmax : float
        maximum mass
    alpha : float
        slope of the power law p(m) = m^-\alpha where alpha > 0

    Returns
    -------
    float or array_like
        m1, m2
    """
    alpha_ = -1*alpha
    u = np.random.rand(N)
    if alpha_ != -1:
        m1 = (u*(mmax**(alpha_+1)-mmin**(alpha_+1)) +
              mmin**(alpha_+1))**(1.0/(alpha_+1))
        print('Powerlaw mass distribution with alpha = ' + str(alpha))
    else:
        m1 = np.exp(u*(np.log(mmax)-np.log(mmin))+np.log(mmin))
        print('Flat in log mass distribution')
    m2 = np.random.uniform(low=5.0, high=m1)
    return m1, m2

def BNS_gaussian_distribution(N, mean=1.35, sigma=0.15):
    """
    Returns p(m1,m2)
    The prior on the mass distribution that follows gaussian for BNSs.

    Parameters
    ----------
    N : integer
        Number of masses sampled
    mean : float
        mean of gaussian dist
    sigma : float
        std of gaussian dist

    Returns
    -------
    float or array_like
        mass1, mass2
    """
    mass1 = []
    mass2 = []
    while len(mass1) < N:
        m1 = np.random.normal(mean, sigma)
        m2 = np.random.normal(mean, sigma)
        if m2 > m1:
            m3 = m2
            m2 = m1
            m1 = m3
        mass1.append(m1)
        mass2.append(m2)
    mass1 = np.array(mass1)
    mass2 = np.array(mass2)
    return mass1, mass2

def BNS_uniform_distribution(N, mmin=1., mmax=3.):
    """
    Returns p(m1,m2)
    The prior on the mass distribution that follows gaussian for BNSs.

    Parameters
    ----------
    N : integer
        Number of masses sampled
    mmin : float
        minimum mass
    mmax : float
        maximum mass

    Returns
    -------
    float or array_like
        mass1, mass2
    """
    mass1 = []
    mass2 = []
    while len(mass1) < N:
        m1 = np.random.uniform(mmin, mmax)
        m2 = np.random.uniform(mmin, mmax)
        if m2 > m1:
            m3 = m2
            m2 = m1
            m1 = m3
        mass1.append(m1)
        mass2.append(m2)
    mass1 = np.array(mass1)
    mass2 = np.array(mass2)
    return mass1, mass2
