import numpy as np
import bilby
import scipy
from ..statistic.priors import uniform_comoving_volume, z_square, z_uniform, uniform_comoving_volume_time_star_rate
from ..cosmology_functions.astrofunction import GW_luminosity_distance, Cm_GW_luminosity_distance

from ..statistic.selection_probability import selection_probability_redshift,selection_probability_redshift_cosiota


class log_CF_likelihood():
    """ This class is used to sample, work and study the Cutler and Flannagan likelihood

    Parameters
    ----------
    det_list: list
        List of detectors objects
    """

    def __init__(self,det_list):
        self.det_list = det_list

    def cpiu(self,cosiota):
        """ This method computes the chipiu var. chipi=0.5*(1+cosiota**2)

        Parameters
        ----------
        cosiota: np.array or float
            cosiota

        Returns
        -------
        np.array
            cpiu
        """

        return 0.5*(1+cosiota**2)

    def compute_theta(self, time, ra, dec, psi,mode_cal=None):
        ''' This method computes the theta matrix (matched filter response matrix)
        in the likelihood. The function also gives back some quantities related to the matrix.

        Parameters
        ----------
        time: np.array
            Time of the merger (tgps)
        ra: np.array
            Right ascension in RAD
        dec: np.array
            Declination in rad
        psi: np.array
            Polarization in radiants
        mode_cal: None or str
            Either None, 'linear' (for linear calibration contribution),'square' for square contribution
            See notebooks with Dimitri.

        Returns
        -------
        theta: np.array
            2x2 matrix corresponding to the [[++,+x,x+,xx]] matched filter response.
        '''
        numerators = []
        numerators_plain = []
        network = self.det_list

        try:
            kappa=self.kappa
        except AttributeError:
            for detector in network:
                psd_det = detector.compute_PSD()
                freq_det = detector.freq
                psd_det = psd_det(freq_det)
                if mode_cal=='linear':
                    numerators.append(freq_det**(-7/3) * psd_det**(-2)/detector.cal_error(freq_det))
                elif mode_cal=='square':
                    numerators.append(freq_det**(-7/3) * psd_det**(-2)/(detector.cal_error(freq_det)*
                                      np.conj(detector.cal_error(freq_det))))
                elif mode_cal is None:
                    numerators.append(freq_det**(-7/3) * psd_det**(-2))

                numerators_plain.append(freq_det**(-7/3) * psd_det**(-2))

            numerators = np.array(numerators,dtype=np.complex_)
            numerators_plain = np.array(numerators_plain,dtype=np.complex_)

            denominator = numerators_plain.mean(0)
            kappa = numerators.sum(1)/denominator.sum()

            self.kappa=kappa

        theta = np.zeros([2,2],dtype=np.complex_)

        for n, detector in enumerate(network):
            fplus, fcross = detector.compute_antenna_patterns(ra,dec,0., time)

            theta[0,0] += kappa[n] * fplus**2
            theta[0,1] += kappa[n] * fplus * fcross
            theta[1,1] += kappa[n] * fcross**2

        theta[1,0] =  theta[0,1]


        return theta

    def compute_epsilon(self, time, ra, dec, psi):
        ''' This method computes the theta matrix (matched filter response matrix)
        in the likelihood. The function also gives back some quantities related to the matrix.

        Parameters
        ----------
        time: np.array
            Time of the merger (tgps)
        ra: np.array
            Right ascension in RAD
        dec: np.array
            Declination in rad
        psi: np.array
            Polarization in radiants

        Returns
        -------
        conditioning: np.array
            Condition number of the theta matrix
        epsilon: np.array
            Degeneracy parameter epsilon
        sigma: np.array
            Detector network response to signal
        delta_psi: np.array
            Rotation for going to dominant polarization frame
        '''
        network = self.det_list
        # Calculate the eigenvalues for each skypoint
        theta = self.compute_theta(time, ra, dec, psi)
        eigvals, _ = np.linalg.eig(theta)
        conditioning = np.linalg.cond(theta)
        eps = (eigvals[0] - eigvals[1])/(eigvals[0] + eigvals[1])
        delta_psi= (eps < 0) * -np.pi/4 + np.arctan(2*theta[0,1]/(-theta[1,1]+theta[0,0]))/4
        epsilon =  np.abs(eps)
        sigma = 0.5 * (eigvals[0] + eigvals[1])

        return conditioning, epsilon, sigma, delta_psi

    def log_pdf_grid(self,distance_array,cosiota_array,signal):
        """Returns the grid and log of the CF on a grid d-cosiota.

        Parameters
        ----------
        distance_array: np.array
            NORMALIZED distance array from which to build
        cosiota_array: np.array
            cosiota array use to build the grid
        signal: dict
            dictionary containing the signal parameters:
                "tgps": Arrival time in GPS
                "dec": Declination in rads
                "RA": right ascension in rads
                "psi": polarization angle in rads

        Returns
        -------
        d_matrix: np.array
            grid matrix on the luminosity distance
        cosiota_matrix: np.array
            grid matrix on the cosiota
        likelihood: np.array
            Log likelihood over the grid
        """

        if len(signal['tgps'])>1:
            raise NameError('You have to inject just 1 signal')

        detectors_network = self.det_list
        tgps = signal['tgps']
        dec = signal['dec']
        ra = signal['RA']
        psi = signal['psi']
        cosiota = np.cos(signal['iota'])
        detected_SNR = signal['SNR_detected']

        chi_piu=(1+cosiota**2)*0.5
        conditioning, epsilon, sigma_d, delta_psi=self.compute_epsilon(tgps,ra,dec,psi)

        rho_0_square=(detected_SNR**2)/(
            sigma_d*(chi_piu**2+cosiota**2+epsilon*(chi_piu**2-cosiota**2)*np.cos(4*(psi+delta_psi))))

        d_matrix, cos_matrix = np.meshgrid(distance_array,cosiota_array)
        likelihood=np.zeros_like(d_matrix,dtype=np.complex_)

        cos4psibar=np.cos(4*(psi-delta_psi))
        chipiu_prime=(1+cos_matrix**2)*0.5
        expon = -rho_0_square*sigma_d*0.5*((1-epsilon*cos4psibar)*(cosiota-cos_matrix/d_matrix)**2.
                                              +(1+epsilon*cos4psibar)*(chi_piu-chipiu_prime/d_matrix)**2.)
        likelihood=expon

        return d_matrix,cos_matrix,np.real(likelihood)


    def log_pdf_point(self,d_values,c_values,signal):
        """Returns the the log of the CF for n-points in d cosiota.

        Parameters
        ----------
        d_values: np.array
            NORMALIZED distance array where to compute the likelihood
        c_values: np.array
            cosiota array where to compute the likelihood
        signal: dict
            dictionary containing the signal parameters:
                "tgps": Arrival time in GPS
                "dec": Declination in rads
                "RA": right ascension in rads
                "psi": polarization angle in rads

        Returns
        -------
        likelihood: np.array
            Values of the likelihood
        """
        if len(signal['tgps'])>1:
            raise NameError('You have to inject just 1 signal')

        detectors_network = self.det_list
        tgps = signal['tgps']
        dec = signal['dec']
        ra = signal['RA']
        psi = signal['psi']
        cosiota = np.cos(signal['iota'])
        detected_SNR = signal['SNR_detected']

        chi_piu=(1+cosiota**2)*0.5
        conditioning, epsilon, sigma_d, delta_psi=self.compute_epsilon(tgps,ra,dec,psi)

        rho_0_square=(detected_SNR**2)/(
            sigma_d*(chi_piu**2+cosiota**2+epsilon*(chi_piu**2-cosiota**2)*np.cos(4*(psi+delta_psi))))

        likelihood=np.zeros_like(d_values,dtype=np.complex_)

        cos4psibar=np.cos(4*(psi-delta_psi))
        chipiu_prime=(1+c_values**2)*0.5
        expon = -rho_0_square*sigma_d*0.5*((1-epsilon*cos4psibar)*(cosiota-c_values/d_values)**2.
                                              +(1+epsilon*cos4psibar)*(chi_piu-chipiu_prime/d_values)**2.)
        likelihood=expon

        return np.real(likelihood)

    def gkde(self,dmin,dmax,signal,cosimin=-2,cosimax=2,Nsample=5000):

        if len(signal['tgps'])>1:
            raise NameError('You have to inject just 1 signal')

        detectors_network = self.det_list
        tgps = signal['tgps']
        dec = signal['dec']
        ra = signal['RA']
        psi = signal['psi']
        cosiota = np.cos(signal['iota'])
        detected_SNR = signal['SNR_detected']

        chi_piu=(1+cosiota**2)*0.5
        conditioning, epsilon, sigma_d, delta_psi=self.compute_epsilon(tgps,ra,dec,psi)

        rho_0_square=(detected_SNR**2)/(
            sigma_d*(chi_piu**2+cosiota**2+epsilon*(chi_piu**2-cosiota**2)*np.cos(4*(psi+delta_psi))))

        d_proxi = np.random.rand(Nsample)*(dmax-dmin)+dmin
        # Note that here we extend cosiota for not having boundaries problems
        cosi_proxi = np.random.rand(Nsample)*(cosimax-cosimin)+cosimin

        logpdf_proxi = self.log_pdf_point(d_proxi,cosi_proxi,signal)
        weights = np.exp(logpdf_proxi)
        weights/=np.max(weights)

        kde_fit=scipy.stats.gaussian_kde(np.vstack([d_proxi,cosi_proxi]),weights=weights)

        return kde_fit

    def sample(self,dmin,dmax,signal,bound=True,cosimin=-2,cosimax=2,Nsample=5000):
        """Returns Nsamples for the CF likelihood. Use a gaussian kernel approximation.

        Parameters
        ----------
        dmin: float
            minimum of the NORMALIZED distance array where to compute the likelihood
        dmax: float
            maximum of the NORMALIZED distance array where to compute the likelihood
        signal: dict
            dictionary containing the signal parameters:
                "tgps": Arrival time in GPS
                "dec": Declination in rads
                "RA": right ascension in rads
                "psi": polarization angle in rads
        Nsample: int
            Number of samples to generate

        Returns
        -------
        samples: np.array
            a 2xNsample array. 1st row: samples on NORMALIZED distance. 2nd row:
            samples on cosiota.
        """

        if len(signal['tgps'])>1:
            raise NameError('You have to inject just 1 signal')

        detectors_network = self.det_list
        tgps = signal['tgps']
        dec = signal['dec']
        ra = signal['RA']
        psi = signal['psi']
        cosiota = np.cos(signal['iota'])
        detected_SNR = signal['SNR_detected']

        chi_piu=(1+cosiota**2)*0.5
        conditioning, epsilon, sigma_d, delta_psi=self.compute_epsilon(tgps,ra,dec,psi)

        rho_0_square=(detected_SNR**2)/(
            sigma_d*(chi_piu**2+cosiota**2+epsilon*(chi_piu**2-cosiota**2)*np.cos(4*(psi+delta_psi))))



        d_proxi = np.random.rand(50000)*(dmax-dmin)+dmin
        # Note that here we extend cosiota for not having boundaries problems
        cosi_proxi = np.random.rand(50000)*(cosimax-cosimin)+cosimin

        logpdf_proxi = self.log_pdf_point(d_proxi,cosi_proxi,signal)
        weights = np.exp(logpdf_proxi)
        weights/=np.max(weights)

        kde_fit=scipy.stats.gaussian_kde(np.vstack([d_proxi,cosi_proxi]),weights=weights)

        new_samps = kde_fit.resample(Nsample)

        if bound:
            cut_index = np.where(np.abs(new_samps[1,:])<=1)[0]
            out_samps=new_samps[:,cut_index]

            while len(out_samps[0,:])!=Nsample:
                N_gen = Nsample-len(out_samps[0,:])
                new_samps = kde_fit.resample(N_gen)
                cut_index = np.where(np.abs(new_samps[1,:])<=1)[0]
                out_samps = np.append(out_samps,new_samps[:,cut_index],axis=1)
        else:
            out_samps=new_samps

        return out_samps

    def sample_iota(self,dmin,dmax,signal,Nsample=5000):
        """Returns Nsamples for the CF likelihood. Use a gaussian kernel approximation.

        Parameters
        ----------
        dmin: float
            minimum of the NORMALIZED distance array where to compute the likelihood
        dmax: float
            maximum of the NORMALIZED distance array where to compute the likelihood
        signal: dict
            dictionary containing the signal parameters:
                "tgps": Arrival time in GPS
                "dec": Declination in rads
                "RA": right ascension in rads
                "psi": polarization angle in rads
        Nsample: int
            Number of samples to generate

        Returns
        -------
        samples: np.array
            a 2xNsample array. 1st row: samples on NORMALIZED distance. 2nd row:
            samples on cosiota.
        """

        if len(signal['tgps'])>1:
            raise NameError('You have to inject just 1 signal')

        detectors_network = self.det_list
        tgps = signal['tgps']
        dec = signal['dec']
        ra = signal['RA']
        psi = signal['psi']
        cosiota = np.cos(signal['iota'])
        detected_SNR = signal['SNR_detected']

        chi_piu=(1+cosiota**2)*0.5
        conditioning, epsilon, sigma_d, delta_psi=self.compute_epsilon(tgps,ra,dec,psi)

        rho_0_square=(detected_SNR**2)/(
            sigma_d*(chi_piu**2+cosiota**2+epsilon*(chi_piu**2-cosiota**2)*np.cos(4*(psi+delta_psi))))



        d_proxi = np.random.rand(10000)*(dmax-dmin)+dmin
        # Note that here we extend cosiota for not having boundaries problems
        central_iota = signal['iota']
        iota_min = signal['iota']-0.5*np.pi
        iota_max = signal['iota']+0.5*np.pi

        iota_proxi = np.random.rand(10000)*(iota_max-iota_min)+iota_min

        logpdf_proxi = self.log_pdf_point(d_proxi,np.cos(iota_proxi),signal)
        weights = np.exp(logpdf_proxi)
        weights/=np.max(weights)

        kde_fit=scipy.stats.gaussian_kde(np.vstack([d_proxi,iota_proxi]),weights=weights)

        new_samps = kde_fit.resample(Nsample)

        return new_samps


def log_gaussian_kde_likelihood(kde_fit,template):
    """ This function returns the log value of a gaussian_kde fit.
    If the returned value is -inf, Then it performs an extrapolation

    Parameters
    ----------
    kde_fit: gaussian kernel fit (scipy)
        Guassian kernel
    template: numpy array Nsample x Dimension
        Template at which compute the log likelihood
    """
    log_prob = kde_fit.logpdf(template)

    for i in range(len(log_prob)):
        if log_prob[i] == -np.inf:
            diff_array = kde_fit.dataset.transpose()-template[i]# Samples x Var
            try:
                distance_val = np.sum(diff_array**2,axis=1)
                index_tem = np.argmin(distance_val)
                new_template = kde_fit.dataset[:,index_tem]
                var_values = np.var(kde_fit.dataset, axis=1)
                log_prob[i] = kde_fit.logpdf(new_template)-np.sum(((diff_array[index_tem,:]**2)/var_values))
            except:
                distance_val = diff_array**2
                index_tem = np.argmin(distance_val)
                new_template = kde_fit.dataset[index_tem]
                var_values = np.var(kde_fit.dataset)
                log_prob[i] = kde_fit.logpdf(new_template[0])-(diff_array[index_tem]**2)/var_values
    return log_prob


class combine_likelihood(bilby.Likelihood):
    """ Bilby likelihood class for the testing cosmo

    Parameters
    ----------
    kernels: list
        List of gaussian kernel fits for the variables
    variables_name: list
        List of strings containing the name of the variables
    """

    def __init__(self, kernels,variables_name):

        self.kernels = kernels
        self.variables_name=variables_name
        self.parameters={}

        for var in variables_name:
            self.parameters[var]=None


    def log_likelihood(self):
        """ Returns the log likelihood
        """

        list_template=[]
        for var in self.variables_name:
            list_template.append(self.parameters[var])

        template=np.array(list_template)
        number_kernels=len(self.kernels)
        log_pdf=0.

        for indx in range(number_kernels):
            log_pdf+=self.kernels[indx].logpdf(template)



        return  log_pdf[0]



class testingGRlikelihood(bilby.Likelihood):
    """ Bilby likelihood class for the testing cosmo

    Parameters
    ----------
    analysis: class
        testingGRcosmo class
    red_prior: str
        Prior to use for redshift: *uniform_comoving_volume*, *z_square*, *z_uniform*
    pop_prior: str
        Prior to use for the population: *uniform_comoving_volume*, *z_square*, *z_uniform*
    """

    def __init__(self, analysis, red_prior='uniform_comoving_volume',
    pop_prior='uniform_comoving_volume',distance_model='GW_luminosity_distance'):

        self.analysis_obj = analysis
        self.red_prior=red_prior
        self.pop_prior=pop_prior
        self.distance_model=distance_model
        super().__init__(parameters={})


        for ss in self.analysis_obj.eq_am.__code__.co_varnames[1:]:
            self.parameters[ss] = None

        for ss in self.analysis_obj.eq_at.__code__.co_varnames[1:]:
            self.parameters[ss] = None

    def log_likelihood(self):
        """ Returns the log likelihood
        """

        am_template = []
        for ss in self.analysis_obj.eq_am.__code__.co_varnames[1:]:
            am_template.append(self.parameters[ss])
        at_template = []
        for ss in self.analysis_obj.eq_at.__code__.co_varnames[1:]:
            at_template.append(self.parameters[ss])

        H0t = self.parameters['H0']
        Om0t = self.parameters['Om0']
        taut = self.parameters['tau']
        freqt = self.parameters['freq_merger']
        z_cosmo = self.parameters['zcosmo']
        z_pec = self.parameters['zpec']

        likelihood_value = self.analysis_obj.likelihood_full_parameters(H0t, Om0t, z_cosmo, at_template, am_template,
         taut, freqt, zpec=z_pec,distance_model=self.distance_model)

        if self.red_prior == 'uniform_comoving_volume':
            likelihood_red=uniform_comoving_volume(z_cosmo,H0=H0t,Om0=Om0t)
        elif self.red_prior == 'z_square':
            likelihood_red=z_square(z_cosmo,Om0=Om0t)
        elif self.red_prior == 'z_uniform':
            likelihood_red=z_uniform(z_cosmo)
        elif self.red_prior == 'uniform_comoving_volume_time_star_rate':
            likelihood_red=uniform_comoving_volume_time_star_rate(z_cosmo,H0=H0t,Om0=Om0t)

        if self.distance_model=='GW_luminosity_distance':
            eq_redshift_to_dgw= lambda z: GW_luminosity_distance(z,
            alpha_M=am_template,lambda_alpha_M=self.analysis_obj.eq_am,Om0=Om0t,H0=H0t)
        elif self.distance_model=='Cm_GW_luminosity_distance':
            eq_redshift_to_dgw= lambda z: Cm_GW_luminosity_distance(z
            ,cm=am_template[0],Om0=Om0t,H0=H0t)

        if self.pop_prior == 'uniform_comoving_volume':
            redshift_pop_prior = lambda z: uniform_comoving_volume(z,H0=H0t,Om0=Om0t)
            selection_prob=selection_probability_redshift(self.analysis_obj.zmax,
            self.analysis_obj.PdetGW,self.analysis_obj.dcrit_Pdet,
            eq_redshift_to_dgw, redshift_pop_prior)
        elif self.pop_prior == 'z_square':
            redshift_pop_prior = lambda z: z_square(z,Om0=Om0t)
            selection_prob=selection_probability_redshift(self.analysis_obj.zmax,
            self.analysis_obj.PdetGW,self.analysis_obj.dcrit_Pdet,
            eq_redshift_to_dgw, redshift_pop_prior)
        elif self.pop_prior == 'z_uniform':
            redshift_pop_prior = lambda z: z_uniform(z)
            selection_prob=selection_probability_redshift(self.analysis_obj.zmax,
            self.analysis_obj.PdetGW,self.analysis_obj.dcrit_Pdet,
            eq_redshift_to_dgw, redshift_pop_prior)
        elif self.pop_prior == 'uniform_comoving_volume_time_star_rate':
            redshift_pop_prior = lambda z: uniform_comoving_volume_time_star_rate(z,H0=H0t,Om0=Om0t)
            selection_prob=selection_probability_redshift(self.analysis_obj.zmax,
            self.analysis_obj.PdetGW,self.analysis_obj.dcrit_Pdet,
            eq_redshift_to_dgw, redshift_pop_prior)
        elif self.pop_prior == 'no':
            selection_prob=0.

        to_ret=likelihood_value+likelihood_red-selection_prob
        return to_ret[0]


class fake_likelihood_EMGW(bilby.Likelihood):
    """ REDSHIFT AND H0 should have same priors
    """

    def __init__(self, analysis, red_prior='uniform_comoving_volume',
    pop_prior='uniform_comoving_volume',prev_analysis=None,log_sel_function_interpolation=None,
    d_l_mode='full'):

        super().__init__(parameters={})
        self.analysis_obj = analysis
        self.red_prior=red_prior
        self.pop_prior=pop_prior
        self.prev_analysis=prev_analysis
        self.log_sel_function_interpolation=log_sel_function_interpolation
        self.d_l_mode=d_l_mode


    def log_likelihood(self):
        """ Returns the log likelihood
        """

        H0 = self.parameters['H0']
        Om0 = self.parameters['Om0']
        zcosmo = self.parameters['zcosmo']
        zpec = self.parameters['zpec']
        cosiota = self.parameters['cosiota']

        log_pos = self.analysis_obj.log_posterior_terms(zcosmo,zpec,H0,Om0,cosiota)



        if self.red_prior == 'uniform_comoving_volume':
            likelihood_red=uniform_comoving_volume(zcosmo,H0=H0,Om0=Om0)
        elif self.red_prior == 'z_square':
            likelihood_red=z_square(zcosmo)
        elif self.red_prior == 'z_uniform':
            likelihood_red=z_uniform(zcosmo)

        if self.prev_analysis is not None:
            prev_anal=self.prev_analysis['log_evidence']+np.log(self.prev_analysis['kernel_posterior_H0'](H0))
        else:
            prev_anal=0

        eq_redshift_to_dgw= lambda z: GW_luminosity_distance(z,Om0=Om0,H0=H0,mode=self.d_l_mode)

        if self.log_sel_function_interpolation is not None:
            selection_prob = self.log_sel_function_interpolation(H0)
        else:

            if self.pop_prior == 'uniform_comoving_volume':
                redshift_pop_prior = lambda z: uniform_comoving_volume(z,H0=H0,Om0=Om0)
                selection_prob = selection_probability_redshift_cosiota(self.analysis_obj.zmax,self.analysis_obj.PdetGW
                ,self.analysis_obj.PdetEM,self.analysis_obj.dcrit_GW,eq_redshift_to_dgw, redshift_pop_prior)
            elif self.pop_prior == 'z_square':
                redshift_pop_prior = lambda z: z_square(z)
                selection_prob = selection_probability_redshift_cosiota(self.analysis_obj.zmax,self.analysis_obj.PdetGW
                ,self.analysis_obj.PdetEM,self.analysis_obj.dcrit_GW,eq_redshift_to_dgw, redshift_pop_prior)
            elif self.pop_prior == 'z_uniform':
                redshift_pop_prior = lambda z: z_uniform(z)
                selection_prob = selection_probability_redshift_cosiota(self.analysis_obj.zmax,self.analysis_obj.PdetGW
                ,self.analysis_obj.PdetEM,self.analysis_obj.dcrit_GW,eq_redshift_to_dgw, redshift_pop_prior)
            elif self.pop_prior == 'no':
                selection_prob=0.

        to_ret=log_pos+likelihood_red-selection_prob+prev_anal
        return to_ret[0]
