import numpy as np
from ..cosmology_functions.astrofunction import E_z
from sys import exit
from scipy.integrate import quad

def uniform_comoving_volume_first_order(redshift_cosmo,H0=67.7329,Om0=0.3):
    q0=0.5*Om0-(1-Om0)
    prob = ((redshift_cosmo**2)/(H0**3))*(1-2*(1+q0)*redshift_cosmo)
    return np.log(prob)

def uniform_comoving_volume(redshift_cosmo,H0=67.7329,Om0=0.3):
    """ This function returns the NATURAL LOG of the prior uniform in comoving volume P(z|H0, Om0)

    Parameters
    ----------
    redshift_cosmo: float
        Cosmological redshift
    H0: float
        Hubble constant in km/Mpc s
    Om0: float
        Dark matter density today

    Returns
    -------
    Natural log of P(z|H0, Om0) : float
    """

    if redshift_cosmo<1e-6:
        exit('Error, use a redshift > than 1e-6')
    func_to_int= lambda z_lam: 1./E_z(z_lam,Om0=Om0)
    d_comoving_square=quad(func_to_int,0,redshift_cosmo)[0]**2.
    return np.log(d_comoving_square/(E_z(redshift_cosmo,Om0=Om0)*(H0**3.)))



def uniform_comoving_volume_time_star_rate(redshift_cosmo,H0=67.7329,Om0=0.3):
    """ This function returns the NATURAL LOG of the prior uniform in comoving volume over (1+z)**1.7
    From https://arxiv.org/pdf/1901.03321.pdf

    Parameters
    ----------
    redshift_cosmo: float
        Cosmological redshift
    H0: float
        Hubble constant in km/Mpc s
    Om0: float
        Dark matter density today

    Returns
    -------
    Natural log of P(z|H0, Om0) : float
    """

    if redshift_cosmo<1e-6:
        exit('Error, use a redshift > than 1e-6')
    func_to_int= lambda z_lam: 1./E_z(z_lam,Om0=Om0)
    d_comoving_square=((1+redshift_cosmo)**1.7)*quad(func_to_int,0,redshift_cosmo)[0]**2.
    return np.log(d_comoving_square/(E_z(redshift_cosmo,Om0=Om0)*(H0**3.)))


def uniform_comoving_volume_low_red(redshift_cosmo,H0=67.7329):
    """ z^2/H0^3

    Parameters
    ----------
    redshift_cosmo: float
        Cosmological redshift
    Returns
    -------
    Natural log of P(z|H0, Om0) : float
    """
    return np.log(redshift_cosmo**2)-np.log(H0**3)


def z_square(redshift_cosmo):
    """ This function returns the NATURAL LOG of the prior z^2

    Parameters
    ----------
    redshift_cosmo: float
        Cosmological redshift
    Returns
    -------
    Natural log of P(z|H0, Om0) : float
    """
    return np.log(redshift_cosmo**2)



def z_uniform(redshift_cosmo):
    """ This function returns the log natural of a uniform in redshift prior

    Parameters
    ----------
    redshift_cosmo: float
        Cosmological redshift

    Returns
    -------
    Natural log of P(z): float
    """

    return 0.
