from sys import exit
from scipy.optimize  import fsolve
import numpy as np
from scipy.integrate import quad, dblquad
from scipy.interpolate import interp1d

def selection_probability_redshift(z_max,PdetGW,dcrit,eq_redshift_to_dgw, redshift_pop_prior):
    """ Returns the Natural LOG of the selection function

    Parameters
    ----------
    z_max: float
        Upperbound in redshift for integrating the selection function.
    PdetGW: interpolation
        GW detection probability in luminosity distance
    dcrit: float
        Where we stop detecting binaries (see the function to compute Pdet)
    eq_redshift_to_dgw: lambda function
        Lambda function to compute dGW from redshift
    redshift_pop_prior: lambda function
        LOG Prior to compute the binary distribution in redshift

    Returns
    -------
    natural log of the selection probability : float
    """
    d_max = eq_redshift_to_dgw(z_max)

    if d_max<dcrit:
        exit('PdetGW is not died at z_max, choose an hiher z_max')

    d_l_gw_fun_solve=lambda z_val:  eq_redshift_to_dgw(z_val)-dcrit
    zcrit=fsolve(d_l_gw_fun_solve,0.01, xtol=0.00001)[0] # Initial solution of 0.01
    func_to_int = lambda z_s: PdetGW(eq_redshift_to_dgw(z_s))*np.exp(redshift_pop_prior(z_s))
    return np.log(quad(func_to_int, 0, z_max,points=[zcrit],epsrel=1e-20)[0])

def selection_probability_redshift_cosiota(z_max,PdetGW,PdetEM,dcrit,eq_redshift_to_dgw, redshift_pop_prior,fast_method=True):
    """ Returns the Natural LOG of the selection function

    Parameters
    ----------
    z_max: float
        Upperbound in redshift for integrating the selection function.
    PdetGW: lambda function
        GW detection probability in (distance_L, cosiota)
    PdetEM: lambda function
        EM Afterglow detection probability in (distance_L, cosiota)
    dcrit: float
        Where we stop detecting binaries (see the function to compute Pdet)
    eq_redshift_to_dgw: lambda function
        Lambda function to compute dGW from redshift
    redshift_pop_prior: lambda function
        LOG Prior to compute the binary distribution in redshift
    fast_method: boolean
        If True perform the double integration on cosiota and d.
        Otherwise marginalizes on cosiota first and then d
    Returns
    -------
    natural log of the selection probability : float
    """
    d_max = eq_redshift_to_dgw(z_max)

    if d_max<dcrit:
        exit('PdetGW is not died at z_max, choose an hiher z_max')

    d_l_gw_fun_solve=lambda z_val:  eq_redshift_to_dgw(z_val)-dcrit
    zcrit=fsolve(d_l_gw_fun_solve,0.01, xtol=0.00001)[0] # Initial solution of 0.01

    if fast_method:

        func_to_int = lambda cosiota, z_s: PdetGW(eq_redshift_to_dgw(z_s),cosiota)*PdetEM(eq_redshift_to_dgw(z_s),cosiota)*np.exp(redshift_pop_prior(z_s))
    else:

        d_array=np.linspace(0.1,dcrit,100)
        marginalized_cosiota=np.zeros_like(d_array)
        for indx in range(len(d_array)):
            to_int = lambda cosiota: PdetGW(d_array[indx],cosiota)*PdetEM(d_array[indx],cosiota)
            marginalized_cosiota[indx]=quad(to_int,-1,1,epsrel=1e-20)[0]

        interpolation_d=interp1d(d_array, marginalized_cosiota,kind='cubic',bounds_error=False,fill_value=0.)

        func_to_int = lambda z_s: interpolation_d(eq_redshift_to_dgw(z_s))*np.exp(redshift_pop_prior(z_s))

        return np.log(quad(func_to_int, 0, z_max,points=[zcrit],epsrel=1e-20)[0])


    return np.log(dblquad(func_to_int, 0, z_max, lambda x: -1, lambda x: 1,epsrel=1e-20)[0])
