import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize as so
import scipy.stats as ss
import seaborn as sns

def plot_2d_kde(ax,xpts,ypts,cmap,levels,shade=True,*args,**kwargs): 

    kde = ss.gaussian_kde(np.row_stack((xpts, ypts)))
    xs = np.linspace(np.min(xpts), np.max(xpts), 100)
    ys = np.linspace(np.min(ypts), np.max(ypts), 100)

    XS, YS = np.meshgrid(xs, ys)
    ZS = np.reshape(kde(np.row_stack((XS.ravel(), YS.ravel()))), XS.shape)

    level_pts = np.row_stack((xpts, ypts))
    lvl = kde(level_pts).flatten()
    zsorted = np.sort(lvl)[::-1]

    lev_contourf=[]
    lev_contourf.append(zsorted[0])

    for lev in levels:
        lev_contourf.append(zsorted[int(lev*len(zsorted))])
    lev_contourf=np.array(lev_contourf[::-1])
    lev_contour=lev_contourf[0:-1:1]

    if shade:
        cset = ax.contourf(XS, YS, ZS, lev_contourf, cmap=cmap, *args, **kwargs)
    else:
        cset=ax.contour(XS, YS, ZS, lev_contour, cmap=cmap, *args, **kwargs)
        fmt={}
        for l,s in zip(cset.levels,levels[::-1]):
            fmt[l]='{:2.0f}%'.format(s*100)
        ax.clabel(cset,cset.levels,fmt=fmt)

        return cset


def max_posterior_intervals(xs, intervals, return_map=True):
    N = xs.shape[0]

    if return_map:
        rho = ss.gaussian_kde(xs)
        xmap = so.brent(lambda x: -rho(x), brack=(np.min(xs), np.median(xs), np.max(xs)))[0]

    xs_sorted = np.sort(xs)

    ilows = []
    ihighs = []
    for i in intervals:
        n = int(round(i*N))

        lowers = xs_sorted[:N-n+1]
        uppers = xs_sorted[n-1:]

        ranges = uppers - lowers

        imin = np.argmin(ranges)

        ilows.append(lowers[imin])
        ihighs.append(uppers[imin])

    if return_map:
        return xmap, zip(ilows, ihighs)
    else:
        return zip(ilows, ihighs)


def plot_1D_PDF(posterior_samples,axes, title=None, levels=[0.683], *args, **kwargs):
    color = kwargs.pop('color', sns.color_palette()[0])

    sns.kdeplot(posterior_samples, ax=axes,color=color, *args, **kwargs)
    axes.legend(loc='upper right')

    if levels is not None:
        bounds = max_posterior_intervals(posterior_samples, levels, return_map=False)

    for i, (l,h) in enumerate(bounds):
        if i % 2 == 0:
            ls = 'dashed'
        else:
            ls = 'dotted'

        u1 = np.percentile(posterior_samples, 84)
        l1 = np.percentile(posterior_samples, 16)

        axes.axvline(l1, color=color, ls=ls)
        axes.axvline(u1, color=color, ls=ls)

    axes.axis(ymin=0)


    #xlabel(r'$H_0$ ($\mathrm{km} \, \mathrm{s}^{-1} \, \mathrm{Mpc}^{-1}$)')
    #ylabel(r'$p\left( H_0 \right)$ ($\mathrm{km}^{-1} \, \mathrm{s} \, \mathrm{Mpc}$)')
    if title is not None:
        axes.set_title(title,fontsize=10)


def kdeplot_credible(xpts, ypts, axes,cmap='viridis', shade=False, shade_lowest=False, levels=None,
                     extra_range=0.25, npts=100, include_lowhigh=True, *args, **kwargs):
    kde = ss.gaussian_kde(np.row_stack((xpts, ypts)))

    xstd = np.std(xpts)
    ystd = np.std(ypts)

    xs = np.linspace(np.min(xpts)-extra_range*xstd, np.max(xpts)+extra_range*xstd, npts)
    ys = np.linspace(np.min(ypts)-extra_range*ystd, np.max(ypts)+extra_range*ystd, npts)

    XS, YS = np.meshgrid(xs, ys)
    ZS = np.reshape(kde(np.row_stack((XS.ravel(), YS.ravel()))), XS.shape)

    level_pts = np.row_stack((xpts, ypts))
    if level_pts.shape[1] > 16384:
        level_pts = level_pts[:, random.choice(level_pts.shape[1], 16384, replace=False)]

    lvl = kde(level_pts)
    ilevels = np.argsort(lvl)[::-1]

    if levels is None:
        ps = np.linspace(0, 1, 11)[1:-1]
    else:
        ps = np.sort(levels)

    inds = (level_pts.shape[1]*ps).round().astype(np.int)

    ls = lvl[ilevels[inds[::-1]]]
    if include_lowhigh:
        ls = np.concatenate(([0.0], ls, [np.max(ZS)]))

    if shade:
        cset = axes.contourf(XS, YS, ZS, ls, cmap=cmap, *args, **kwargs)
        if not shade_lowest:
            cset.collections[0].set_alpha(0)
        return cset
    else:
        return axes.contour(XS, YS, ZS, ls, cmap=cmap, *args, **kwargs)
